# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utility methods for scheduling."""

import collections
import math
import re

import os_resource_classes as orc
from oslo_log import log as logging
from urllib import parse

from zun.common import exception
import zun.conf
from zun import objects


LOG = logging.getLogger(__name__)
CONF = zun.conf.CONF


class ResourceRequest(object):
    """Presents a granular resource request via RequestGroup instances."""
    # extra_specs-specific consts
    XS_RES_PREFIX = 'resources'
    XS_TRAIT_PREFIX = 'trait'
    # Regex patterns for numbered or un-numbered resources/trait keys
    XS_KEYPAT = re.compile(r"^(%s)([1-9][0-9]*)?:(.*)$" %
                           '|'.join((XS_RES_PREFIX, XS_TRAIT_PREFIX)))

    def __init__(self):
        # { ident: RequestGroup }
        self._rg_by_id = {}
        self._group_policy = None
        # Default to the configured limit but _limit can be
        # set to None to indicate "no limit".
        self._limit = CONF.scheduler.max_placement_results

    def __str__(self):
        return ', '.join(sorted(
            list(str(rg) for rg in list(self._rg_by_id.values()))))

    @property
    def group_policy(self):
        return self._group_policy

    @group_policy.setter
    def group_policy(self, value):
        self._group_policy = value

    def get_request_group(self, ident):
        if ident not in self._rg_by_id:
            rq_grp = objects.RequestGroup(use_same_provider=bool(ident))
            self._rg_by_id[ident] = rq_grp
        return self._rg_by_id[ident]

    def add_request_group(self, request_group):
        """Inserts the existing group with a unique integer id

        This function can ensure unique ids by using bigger
        ids than the maximum of existing ids.

        :param request_group: the RequestGroup to be added
        """
        # NOTE(gibi) [0] just here to always have a defined maximum
        group_idents = [0] + [int(ident) for ident in self._rg_by_id if ident]
        ident = max(group_idents) + 1
        self._rg_by_id[ident] = request_group

    def _add_resource(self, groupid, rclass, amount):
        # Validate the class.
        if not (rclass.startswith(orc.CUSTOM_NAMESPACE) or
                rclass in orc.STANDARDS):
            LOG.warning(
                "Received an invalid ResourceClass '%(key)s' in extra_specs.",
                {"key": rclass})
            return
        # val represents the amount.  Convert to int, or warn and skip.
        try:
            amount = int(amount)
            if amount < 0:
                raise ValueError()
        except ValueError:
            LOG.warning(
                "Resource amounts must be nonnegative integers. Received "
                "'%(val)s' for key resources%(groupid)s.",
                {"groupid": groupid or '', "val": amount})
            return
        self.get_request_group(groupid).resources[rclass] = amount

    def _add_trait(self, groupid, trait_name, trait_type):
        # Currently the only valid values for a trait entry are 'required'
        # and 'forbidden'
        trait_vals = ('required', 'forbidden')
        if trait_type == 'required':
            self.get_request_group(groupid).required_traits.add(trait_name)
        elif trait_type == 'forbidden':
            self.get_request_group(groupid).forbidden_traits.add(trait_name)
        else:
            LOG.warning(
                "Only (%(tvals)s) traits are supported. Received '%(val)s' "
                "for key trait%(groupid)s.",
                {"tvals": ', '.join(trait_vals), "groupid": groupid or '',
                 "val": trait_type})
        return

    def _add_group_policy(self, policy):
        # The only valid values for group_policy are 'none' and 'isolate'.
        if policy not in ('none', 'isolate'):
            LOG.warning(
                "Invalid group_policy '%s'. Valid values are 'none' and "
                "'isolate'.", policy)
            return
        self._group_policy = policy

    @classmethod
    def from_extra_specs(cls, extra_specs, req=None):
        """Processes resources and traits in numbered groupings in extra_specs.

        Examines extra_specs for items of the following forms:
            "resources:$RESOURCE_CLASS": $AMOUNT
            "resources$N:$RESOURCE_CLASS": $AMOUNT
            "trait:$TRAIT_NAME": "required"
            "trait$N:$TRAIT_NAME": "required"

        Does *not* yet handle member_of[$N].

        :param extra_specs: The extra_specs dict.
        :param req: the ResourceRequest object to add the requirements to or
               None to create a new ResourceRequest
        :return: A ResourceRequest object representing the resources and
                 required traits in the extra_specs.
        """
        # TODO(efried): Handle member_of[$N], which will need to be reconciled
        # with destination.aggregates handling in resources_from_request_spec

        if req is not None:
            ret = req
        else:
            ret = cls()

        for key, val in extra_specs.items():
            if key == 'group_policy':
                ret._add_group_policy(val)
                continue

            match = cls.XS_KEYPAT.match(key)
            if not match:
                continue

            # 'prefix' is 'resources' or 'trait'
            # 'suffix' is $N or None
            # 'name' is either the resource class name or the trait name.
            prefix, suffix, name = match.groups()

            # Process "resources[$N]"
            if prefix == cls.XS_RES_PREFIX:
                ret._add_resource(suffix, name, val)

            # Process "trait[$N]"
            elif prefix == cls.XS_TRAIT_PREFIX:
                ret._add_trait(suffix, name, val)

        return ret

    def resource_groups(self):
        for rg in self._rg_by_id.values():
            yield rg.resources

    def get_num_of_numbered_groups(self):
        return len([ident for ident in self._rg_by_id.keys()
                    if ident is not None])

    def merged_resources(self, resources=None):
        """Returns a merge of {resource_class: amount} for all resource groups.

        Amounts of the same resource class from different groups are added
        together.

        :param resources: A flat dict of {resource_class: amount}.  If
                          specified, the resources therein are folded
                          into the return dict, such that any resource
                          in resources is included only if that
                          resource class does not exist elsewhere in the
                          merged ResourceRequest.
        :return: A dict of the form {resource_class: amount}
        """
        ret = collections.defaultdict(lambda: 0)
        for resource_dict in self.resource_groups():
            for resource_class, amount in resource_dict.items():
                ret[resource_class] += amount
        if resources:
            for resource_class, amount in resources.items():
                # If it's in there - even if zero - ignore the one from the
                # flavor.
                if resource_class not in ret:
                    ret[resource_class] = amount
            # Now strip zeros.  This has to be done after the above - we can't
            # use strip_zeros :(
            ret = {rc: amt for rc, amt in ret.items() if amt}
        return dict(ret)

    def _clean_empties(self):
        """Get rid of any empty ResourceGroup instances."""
        for ident, rg in list(self._rg_by_id.items()):
            if not any((rg.resources, rg.required_traits,
                        rg.forbidden_traits)):
                self._rg_by_id.pop(ident)

    def strip_zeros(self):
        """Remove any resources whose amounts are zero."""
        for resource_dict in self.resource_groups():
            for rclass in list(resource_dict):
                if resource_dict[rclass] == 0:
                    resource_dict.pop(rclass)
        self._clean_empties()

    def to_querystring(self):
        """Produce a querystring of the form expected by
        GET /allocation_candidates.
        """
        # TODO(gibi): We have a RequestGroup OVO so we can move this to that
        # class as a member function.
        # NOTE(efried): The sorting herein is not necessary for the API; it is
        # to make testing easier and logging/debugging predictable.
        def to_queryparams(request_group, suffix):
            res = request_group.resources
            required_traits = request_group.required_traits
            forbidden_traits = request_group.forbidden_traits
            aggregates = request_group.aggregates
            in_tree = request_group.in_tree

            resource_query = ",".join(
                sorted("%s:%s" % (rc, amount)
                       for (rc, amount) in res.items()))
            qs_params = [('resources%s' % suffix, resource_query)]

            # Assemble required and forbidden traits, allowing for either/both
            # to be empty.
            required_val = ','.join(
                sorted(required_traits) +
                ['!%s' % ft for ft in sorted(forbidden_traits)])
            if required_val:
                qs_params.append(('required%s' % suffix, required_val))
            if aggregates:
                aggs = []
                # member_ofN is a list of lists.  We need a tuple of
                # ('member_ofN', 'in:uuid,uuid,...') for each inner list.
                for agglist in aggregates:
                    aggs.append(('member_of%s' % suffix,
                                 'in:' + ','.join(sorted(agglist))))
                qs_params.extend(sorted(aggs))
            if in_tree:
                qs_params.append(('in_tree%s' % suffix, in_tree))
            return qs_params

        if self._limit is not None:
            qparams = [('limit', self._limit)]
        else:
            qparams = []
        if self._group_policy is not None:
            qparams.append(('group_policy', self._group_policy))

        for ident, rg in self._rg_by_id.items():
            # [('resourcesN', 'rclass:amount,rclass:amount,...'),
            #  ('requiredN', 'trait_name,!trait_name,...'),
            #  ('member_ofN', 'in:uuid,uuid,...'),
            #  ('member_ofN', 'in:uuid,uuid,...')]
            qparams.extend(to_queryparams(rg, ident or ''))

        return parse.urlencode(sorted(qparams))


def resources_from_request_spec(ctxt, container_obj, extra_specs):
    """Given a Container object, returns a ResourceRequest of the resources,
    traits, and aggregates it represents.
    :param ctxt: The request context.
    :param container_obj: A Container object.
    :return: A ResourceRequest object.
    :raises NoValidHost: If the specified host/node is not found in the DB.
    """
    cpu = container_obj.cpu if container_obj.cpu else CONF.default_cpu
    # NOTE(hongbin): Container is allowed to take partial core (i.e. 0.1)
    # but placement doesn't support it. Therefore, we take the ceil of
    # the number.
    cpu = int(math.ceil(cpu))
    # NOTE(hongbin): If cpu is 0, claim 1 core in placement because placement
    # doesn't support cpu as 0.
    cpu = cpu if cpu > 1 else 1
    memory = int(container_obj.memory) if container_obj.memory else \
        CONF.default_memory
    # NOTE(hongbin): If memory is 0, claim 1 MB in placement because placement
    # doesn't support memory as 0.
    memory = memory if memory > 1 else 1

    container_resources = {
        orc.VCPU: cpu,
        orc.MEMORY_MB: memory,
    }

    if container_obj.disk and container_obj.disk != 0:
        container_resources[orc.DISK_GB] = container_obj.disk

    # Process extra_specs
    if extra_specs:
        res_req = ResourceRequest.from_extra_specs(extra_specs)
        # If any of the three standard resources above was explicitly given in
        # the extra_specs - in any group - we need to replace it, or delete it
        # if it was given as zero.  We'll do this by grabbing a merged version
        # of the ResourceRequest resources and removing matching items from the
        # container_resources.
        container_resources = {rclass: amt
                               for rclass, amt in container_resources.items()
                               if rclass not in res_req.merged_resources()}
        # Now we don't need (or want) any remaining zero entries - remove them.
        res_req.strip_zeros()

        numbered_groups = res_req.get_num_of_numbered_groups()
    else:
        # Start with an empty one
        res_req = ResourceRequest()
        numbered_groups = 0

    # Add the (remaining) items from the container_resources to the
    # sharing group
    for rclass, amount in container_resources.items():
        res_req.get_request_group(None).resources[rclass] = amount

    requested_resources = extra_specs.get('requested_resources', [])
    for group in requested_resources:
        res_req.add_request_group(group)

    target_host = extra_specs.get('requested_host')
    if target_host:
        nodes = objects.ComputeNode.list(
            ctxt, filters={'hostname': target_host})
        if not nodes:
            reason = (_('No such host - host: %(host)s ') %
                      {'host': target_host})
            raise exception.NoValidHost(reason=reason)
        if len(nodes) == 1:
            grp = res_req.get_request_group(None)
            grp.in_tree = nodes[0].rp_uuid
        else:
            # Multiple nodes are found when a target host is specified
            # without a specific node. Since placement doesn't support
            # multiple uuids in the `in_tree` queryparam, what we can do here
            # is to remove the limit from the `GET /a_c` query to prevent
            # the found nodes from being filtered out in placement.
            res_req._limit = None

    # Don't limit allocation candidates when using affinity/anti-affinity.
    if (extra_specs.get('hints') and any(
            key in ['group', 'same_host', 'different_host']
            for key in extra_specs.get('hints'))):
        res_req._limit = None

    if res_req.get_num_of_numbered_groups() >= 2 and not res_req.group_policy:
        LOG.warning(
            "There is more than one numbered request group in the "
            "allocation candidate query but the container did not specify "
            "any group policy. This query would fail in placement due to "
            "the missing group policy. If you specified more than one "
            "numbered request group in the extra_spec then you need to "
            "specify the group policy in the extra_spec. If it is OK "
            "to let these groups be satisfied by overlapping resource "
            "providers then use 'group_policy': 'none'. If you want each "
            "group to be satisfied from a separate resource provider then "
            "use 'group_policy': 'isolate'.")

        if numbered_groups <= 1:
            LOG.info(
                "At least one numbered request group is defined outside of "
                "the container (e.g. in a port that has a QoS minimum "
                "bandwidth policy rule attached) but the flavor did not "
                "specify any group policy. To avoid the placement failure "
                "nova defaults the group policy to 'none'.")
            res_req.group_policy = 'none'

    return res_req


def claim_resources(ctx, client, container, alloc_req,
                    allocation_request_version=None):
    """Given a container and the
    allocation_request JSON object returned from Placement, attempt to claim
    resources for the container in the placement API. Returns True if the claim
    process was successful, False otherwise.
    :param ctx: The RequestContext object
    :param client: The scheduler client to use for making the claim call
    :param container: The consuming container
    :param alloc_req: The allocation_request received from placement for the
                      resources we want to claim against the chosen host. The
                      allocation_request satisfies the original request for
                      resources and can be supplied as-is (along with the
                      project and user ID to the placement API's PUT
                      /allocations/{consumer_uuid} call to claim resources for
                      the container
    :param allocation_request_version: The microversion used to request the
                                       allocations.
    """
    LOG.debug("Attempting to claim resources in the placement API for "
              "container %s", container.uuid)

    project_id = container.project_id
    user_id = container.user_id
    container_uuid = container.uuid

    # NOTE(gibi): this could raise AllocationUpdateFailed which means there is
    # a serious issue with the container_uuid as a consumer. Every caller of
    # utils.claim_resources() assumes that container_uuid will be a new
    # consumer and therefore we passing None as expected consumer_generation to
    # reportclient.claim_resources() here. If the claim fails
    # due to consumer generation conflict, which in this case means the
    # consumer is not new, then we let the AllocationUpdateFailed propagate and
    # fail the build / migrate as the instance is in inconsistent state.
    return client.claim_resources(
        ctx, container_uuid, alloc_req, project_id, user_id,
        allocation_request_version=allocation_request_version,
        consumer_generation=None)



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock
from unittest.mock import patch

from neutronclient.common import exceptions as n_exc
from oslo_utils import uuidutils
from webtest.app import AppError

from zun.common import exception
from zun import objects
from zun.tests.unit.api import base as api_base
from zun.tests.unit.db import utils
from zun.tests.unit.objects import utils as obj_utils


class TestContainerController(api_base.FunctionalTest):
    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_run_container(self, mock_search, mock_container_create,
                           mock_neutron_get_network):
        mock_container_create.side_effect = lambda x, y, **z: y

        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"}}')

        response = self.post('/v1/containers?run=true',
                             params=params,
                             content_type='application/json')

        self.assertEqual(202, response.status_int)
        self.assertTrue(mock_container_create.called)
        self.assertTrue(mock_container_create.call_args[1]['run'] is True)
        mock_neutron_get_network.assert_called_once()

    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_run_container_wrong_run_value(self, mock_search,
                                           mock_container_create):
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"}}')
        with self.assertRaisesRegex(AppError,
                                    "Invalid input for query parameters"):
            self.post('/v1/containers?run=xyz', params=params,
                      content_type='application/json')

    @patch('zun.compute.api.API.container_create')
    def test_run_container_wrong_memory_value(self, mock_container_create):
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "3",'
                  '"environment": {"key1": "val1", "key2": "val2"}}')
        with self.assertRaisesRegex(AppError,
                                    "Invalid input for field .*memory"):
            self.post('/v1/containers?run=true', params=params,
                      content_type='application/json')

    @patch('zun.compute.api.API.container_create')
    def test_run_container_wrong_cpu_value(self, mock_container_create):
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512", "cpu": "100",'
                  '"environment": {"key1": "val1", "key2": "val2"}}')
        with self.assertRaisesRegex(AppError,
                                    "Invalid input for field .*cpu"):
            self.post('/v1/containers?run=true', params=params,
                      content_type='application/json')

    @patch('zun.compute.api.API.container_create')
    def test_run_container_wrong_disk_value(self, mock_container_create):
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512", "disk": "0",'
                  '"environment": {"key1": "val1", "key2": "val2"}}')
        with self.assertRaisesRegex(AppError,
                                    "Invalid input for field .*disk"):
            self.post('/v1/containers?run=true', params=params,
                      content_type='application/json')

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_create')
    def test_run_container_wrong_exposed_ports(
            self, mock_container_create, mock_mock_neutron_get_network):
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"exposed_ports": {"foo": {}}}')
        with self.assertRaisesRegex(AppError,
                                    "value foo is invalid as publish port"):
            self.post('/v1/containers?run=true', params=params,
                      content_type='application/json')

        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"exposed_ports": {"80/foo": {}}}')
        with self.assertRaisesRegex(AppError,
                                    "value foo is an invalid protocol"):
            self.post('/v1/containers?run=true', params=params,
                      content_type='application/json')

    def test_run_container_runtime_wrong_api_version(self):
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": "env", "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"},'
                  '"runtime": "runc"}')
        headers = {"OpenStack-API-Version": "container 1.4"}
        with self.assertRaisesRegex(AppError,
                                    "Invalid param runtime"):
            self.post('/v1/containers?run=true',
                      params=params, content_type='application/json',
                      headers=headers)

    def test_run_container_runtime_wrong_value(self):
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"},'
                  '"runtime": 1234}')
        with self.assertRaisesRegex(AppError,
                                    "Invalid input for field"):
            self.post('/v1/containers?run=true',
                      params=params, content_type='application/json')

    def test_run_container_with_hostname_wrong_api_version(self):
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": "env", "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"},'
                  '"hostname": "testhost"}')
        headers = {"OpenStack-API-Version": "container 1.7"}
        with self.assertRaisesRegex(AppError,
                                    "Invalid param hostname"):
            self.post('/v1/containers?run=true',
                      params=params, content_type='application/json',
                      headers=headers)

    def test_run_container_healthcheck_wrong_value(self):
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"},'
                  '"healthcheck": "test"}')
        with self.assertRaisesRegex(AppError,
                                    "Invalid input for field"):
            self.post('/v1/containers?run=true',
                      params=params, content_type='application/json')

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_run_container_with_false(self, mock_search,
                                      mock_container_create,
                                      mock_neutron_get_network):
        mock_container_create.side_effect = lambda x, y, **z: y

        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"}}')
        response = self.post('/v1/containers?run=false',
                             params=params,
                             content_type='application/json')
        self.assertEqual(202, response.status_int)
        self.assertTrue(mock_container_create.called)
        self.assertTrue(mock_container_create.call_args[1]['run'] is False)
        mock_neutron_get_network.assert_called_once()

    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_run_container_with_wrong(self, mock_search,
                                      mock_container_create):
        mock_container_create.side_effect = exception.InvalidValue
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"}}')
        self.assertRaises(AppError, self.post, '/v1/containers?run=wrong',
                          params=params, content_type='application/json')
        self.assertTrue(mock_container_create.assert_not_called)

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_create_container(self, mock_search, mock_container_create,
                              mock_neutron_get_network):
        mock_container_create.side_effect = lambda x, y, **z: y
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"}}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')

        self.assertEqual(202, response.status_int)
        self.assertNotIn('host', response.json)
        self.assertTrue(mock_container_create.called)
        self.assertTrue(mock_container_create.call_args[1]['run'] is False)
        mock_neutron_get_network.assert_called_once()

    @patch('zun.common.context.RequestContext.can')
    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_by_admin(
            self, mock_search, mock_container_create, mock_neutron_get_network,
            mock_can):
        mock_container_create.side_effect = lambda x, y, **z: y
        mock_can.return_value = True
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"}}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')

        self.assertEqual(202, response.status_int)
        self.assertIn('host', response.json)
        self.assertTrue(mock_container_create.called)
        self.assertTrue(mock_container_create.call_args[1]['run'] is False)
        mock_neutron_get_network.assert_called_once()

    @patch('zun.compute.api.API.container_create')
    def test_create_container_image_not_specified(self, mock_container_create):

        params = ('{"name": "MyDocker",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"}}')
        with self.assertRaisesRegex(AppError,
                                    "is a required property"):
            self.post('/v1/containers/',
                      params=params,
                      content_type='application/json')
        self.assertTrue(mock_container_create.assert_not_called)

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_set_project_id_and_user_id(
            self, mock_search, mock_container_create,
            mock_neutron_get_network):
        def _create_side_effect(cnxt, container, **kwargs):
            self.assertEqual(self.context.project_id, container.project_id)
            self.assertEqual(self.context.user_id, container.user_id)
            return container
        mock_container_create.side_effect = _create_side_effect

        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"}}')
        self.post('/v1/containers/',
                  params=params,
                  content_type='application/json')
        mock_neutron_get_network.assert_called_once()

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_resp_has_status_reason(self, mock_search,
                                                     mock_container_create,
                                                     mock_neutron_get_network):
        mock_container_create.side_effect = lambda x, y, **z: y
        # Create a container with a command
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"}}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(202, response.status_int)
        self.assertIn('status_reason', response.json.keys())
        mock_neutron_get_network.assert_called_once()

    @patch('zun.common.policy.enforce')
    @patch('zun.network.neutron.NeutronAPI.get_neutron_network')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.container_delete')
    @patch('zun.volume.cinder_api.CinderAPI.search_volume')
    @patch('zun.volume.cinder_api.CinderAPI.ensure_volume_usable')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_with_full_params(self, mock_search,
                                               mock_ensure_volume_usable,
                                               mock_search_volume,
                                               mock_container_delete,
                                               mock_container_create,
                                               mock_neutron_get_network,
                                               mock_policy):
        mock_policy.return_value = True
        mock_container_create.side_effect = lambda x, y, **z: y
        fake_network = {'id': 'foo'}
        mock_neutron_get_network.return_value = fake_network
        fake_volume_id = 'fakevolid'
        fake_volume = mock.Mock(id=fake_volume_id)
        mock_search_volume.return_value = fake_volume
        # Create a container with a command
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"},'
                  '"runtime": "runc", "hostname": "testhost",'
                  '"disk": 20, "restart_policy": {"Name": "no"},'
                  '"nets": [{"network": "testpublicnet"}],'
                  '"exposed_ports": {"80/tcp": {}, "22": {}},'
                  '"mounts": [{"source": "s", "destination": "d"}]}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(202, response.status_int)
        response = self.get('/v1/containers/')
        self.assertEqual(200, response.status_int)
        self.assertEqual(2, len(response.json))
        c = response.json['containers'][0]
        self.assertIsNotNone(c.get('uuid'))
        self.assertEqual('MyDocker', c.get('name'))
        self.assertEqual(["env"], c.get('command'))
        self.assertEqual('512', c.get('memory'))
        self.assertEqual({"key1": "val1", "key2": "val2"},
                         c.get('environment'))
        self.assertEqual('testhost', c.get('hostname'))
        self.assertEqual(20, c.get('disk'))
        self.assertEqual({"Name": "no", "MaximumRetryCount": "0"},
                         c.get('restart_policy'))
        self.assertNotIn('host', c)
        requested_networks = \
            mock_container_create.call_args[1]['requested_networks']
        self.assertEqual(1, len(requested_networks))
        self.assertEqual(fake_network['id'], requested_networks[0]['network'])
        requested_volumes = \
            mock_container_create.call_args[1]['requested_volumes']
        self.assertEqual(1, len(requested_volumes))
        self.assertEqual(
            fake_volume_id,
            next(iter(requested_volumes.values()))[0].cinder_volume_id)
        exposed_ports = mock_container_create.call_args[0][1].exposed_ports
        self.assertEqual(2, len(exposed_ports))
        self.assertIn("80/tcp", exposed_ports)
        self.assertIn("22/tcp", exposed_ports)

        # Delete the container we created
        def side_effect(*args, **kwargs):
            (ctx, cnt, force) = args
            cnt.destroy(ctx)

        mock_container_delete.side_effect = side_effect
        response = self.delete(
            '/v1/containers/%s?force=True' % c.get('uuid'))
        self.assertEqual(204, response.status_int)

        response = self.get('/v1/containers/')
        self.assertEqual(200, response.status_int)
        c = response.json['containers']
        self.assertEqual(0, len(c))
        self.assertTrue(mock_container_create.called)
        mock_neutron_get_network.assert_called_once()

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_with_minimum_params(self,
                                                  mock_search,
                                                  mock_container_create,
                                                  mock_neutron_get_network):
        mock_container_create.side_effect = lambda x, y, **z: y
        fake_network = {'id': 'foo'}
        mock_neutron_get_network.return_value = fake_network
        # Create a container with a command
        params = ('{"image": "ubuntu"}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(202, response.status_int)
        response = self.get('/v1/containers/')
        self.assertEqual(200, response.status_int)
        self.assertEqual(2, len(response.json))
        c = response.json['containers'][0]
        self.assertIsNotNone(c.get('uuid'))
        self.assertIsNotNone(c.get('name'))
        self.assertFalse(c.get('command'))
        self.assertEqual('512', c.get('memory'))
        self.assertEqual(1.0, c.get('cpu'))
        # TODO(kiennt): Uncomment it when bug [1] be resolved.
        #               At this time, limit disk size feature will be ready.
        # [1] https://bugs.launchpad.net/zun/+bug/1746401
        # self.assertEqual(10, c.get('disk'))
        self.assertEqual({}, c.get('environment'))
        self.assertIsNone(c.get('runtime'))
        self.assertIsNone(c.get('hostname'))
        self.assertEqual({}, c.get('restart_policy'))
        self.assertNotIn('host', c)
        mock_neutron_get_network.assert_called_once()
        requested_networks = \
            mock_container_create.call_args[1]['requested_networks']
        self.assertEqual(1, len(requested_networks))
        self.assertEqual(fake_network['id'], requested_networks[0]['network'])

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_with_availability_zone(
            self, mock_search, mock_container_create,
            mock_neutron_get_network):
        mock_container_create.side_effect = lambda x, y, **z: y
        fake_network = {'id': 'foo'}
        mock_neutron_get_network.return_value = fake_network
        # Create a container with a command
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"],'
                  '"availability_zone": "test-az"}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(202, response.status_int)
        response = self.get('/v1/containers/')
        self.assertEqual(200, response.status_int)
        self.assertEqual(2, len(response.json))
        c = response.json['containers'][0]
        self.assertIsNotNone(c.get('uuid'))
        self.assertEqual('MyDocker', c.get('name'))
        self.assertEqual(["env"], c.get('command'))
        self.assertEqual('512', c.get('memory'))
        self.assertEqual(1.0, c.get('cpu'))
        # TODO(kiennt): Uncomment it when bug [1] be resolved.
        #               At this time, limit disk size feature will be ready.
        # [1] https://bugs.launchpad.net/zun/+bug/1746401
        # self.assertEqual(10, c.get('disk'))
        self.assertEqual({}, c.get('environment'))
        self.assertNotIn('host', c)
        mock_neutron_get_network.assert_called_once()
        extra_spec = \
            mock_container_create.call_args[1]['extra_spec']
        self.assertEqual('test-az', extra_spec['availability_zone'])

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_with_restart_policy_no_retry_0(
            self,
            mock_search,
            mock_container_create,
            mock_neutron_get_network):
        mock_container_create.side_effect = lambda x, y, **z: y
        fake_network = {'id': 'foo'}
        mock_neutron_get_network.return_value = fake_network
        # Create a container with a command
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"restart_policy": {"Name": "no",'
                  '"MaximumRetryCount": "0"}}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(202, response.status_int)
        response = self.get('/v1/containers/')
        self.assertEqual(200, response.status_int)
        self.assertEqual(2, len(response.json))
        c = response.json['containers'][0]
        self.assertIsNotNone(c.get('uuid'))
        self.assertEqual('MyDocker', c.get('name'))
        self.assertEqual(["env"], c.get('command'))
        self.assertEqual('512', c.get('memory'))
        self.assertEqual({"Name": "no", "MaximumRetryCount": "0"},
                         c.get('restart_policy'))
        self.assertNotIn('host', c)
        mock_neutron_get_network.assert_called_once()
        requested_networks = \
            mock_container_create.call_args[1]['requested_networks']
        self.assertEqual(1, len(requested_networks))
        self.assertEqual(fake_network['id'], requested_networks[0]['network'])

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_with_restart_policy_no_retry_6(
            self,
            mock_search,
            mock_container_create,
            mock_neutron_get_network):
        mock_container_create.side_effect = lambda x, y, **z: y
        fake_network = {'id': 'foo'}
        mock_neutron_get_network.return_value = fake_network
        # Create a container with a command
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"restart_policy": {"Name": "no",'
                  '"MaximumRetryCount": "6"}}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(202, response.status_int)
        response = self.get('/v1/containers/')
        self.assertEqual(200, response.status_int)
        self.assertEqual(2, len(response.json))
        c = response.json['containers'][0]
        self.assertIsNotNone(c.get('uuid'))
        self.assertEqual('MyDocker', c.get('name'))
        self.assertEqual(["env"], c.get('command'))
        self.assertEqual('512', c.get('memory'))
        self.assertEqual({"Name": "no", "MaximumRetryCount": "0"},
                         c.get('restart_policy'))
        self.assertNotIn('host', c)
        mock_neutron_get_network.assert_called_once()
        requested_networks = \
            mock_container_create.call_args[1]['requested_networks']
        self.assertEqual(1, len(requested_networks))
        self.assertEqual(fake_network['id'], requested_networks[0]['network'])

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_with_restart_policy_unless_stopped(
            self,
            mock_search,
            mock_container_create,
            mock_neutron_get_network):
        mock_container_create.side_effect = lambda x, y, **z: y
        fake_network = {'id': 'foo'}
        mock_neutron_get_network.return_value = fake_network
        # Create a container with a command
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"restart_policy": {"Name": "unless-stopped",'
                  '"MaximumRetryCount": "0"}}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(202, response.status_int)
        response = self.get('/v1/containers/')
        self.assertEqual(200, response.status_int)
        self.assertEqual(2, len(response.json))
        c = response.json['containers'][0]
        self.assertIsNotNone(c.get('uuid'))
        self.assertEqual('MyDocker', c.get('name'))
        self.assertEqual(["env"], c.get('command'))
        self.assertEqual('512', c.get('memory'))
        self.assertEqual({"Name": "unless-stopped", "MaximumRetryCount": "0"},
                         c.get('restart_policy'))
        self.assertNotIn('host', c)
        mock_neutron_get_network.assert_called_once()
        requested_networks = \
            mock_container_create.call_args[1]['requested_networks']
        self.assertEqual(1, len(requested_networks))
        self.assertEqual(fake_network['id'], requested_networks[0]['network'])

    @patch('zun.common.policy.enforce')
    @patch('neutronclient.v2_0.client.Client.show_port')
    @patch('zun.network.neutron.NeutronAPI.get_neutron_network')
    @patch('zun.network.neutron.NeutronAPI.get_neutron_port')
    @patch('zun.network.neutron.NeutronAPI.ensure_neutron_port_usable')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.container_delete')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_with_requested_neutron_port(
            self, mock_search, mock_container_delete, mock_container_create,
            mock_ensure_port_usable, mock_get_port,
            mock_get_network, mock_show_port, mock_policy):
        mock_policy.return_value = True
        mock_container_create.side_effect = lambda x, y, **z: y
        fake_port = {'network_id': 'foo', 'id': 'bar'}
        fake_private_network = {'router:external': False, 'shared': False}
        mock_get_port.return_value = fake_port
        mock_get_network.return_value = fake_private_network
        mock_show_port.return_value = {'port': fake_port}
        # Create a container with a command
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"},'
                  '"nets": [{"port": "testport"}]}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(202, response.status_int)
        response = self.get('/v1/containers/')
        self.assertEqual(200, response.status_int)
        self.assertEqual(2, len(response.json))
        c = response.json['containers'][0]
        self.assertIsNotNone(c.get('uuid'))
        self.assertEqual('MyDocker', c.get('name'))
        self.assertEqual(["env"], c.get('command'))
        self.assertEqual('512', c.get('memory'))
        self.assertEqual({"key1": "val1", "key2": "val2"},
                         c.get('environment'))
        self.assertNotIn('host', c)
        requested_networks = \
            mock_container_create.call_args[1]['requested_networks']
        self.assertEqual(1, len(requested_networks))
        self.assertEqual(fake_port['network_id'],
                         requested_networks[0]['network'])
        self.assertEqual(fake_port['id'], requested_networks[0]['port'])
        self.assertTrue(requested_networks[0]['preserve_on_delete'])

        def side_effect(*args, **kwargs):
            (ctx, cnt, force) = args
            cnt.destroy(ctx)

        # Delete the container we created
        mock_container_delete.side_effect = side_effect
        response = self.delete(
            '/v1/containers/%s?force=True' % c.get('uuid'))
        self.assertEqual(204, response.status_int)

        response = self.get('/v1/containers/')
        self.assertEqual(200, response.status_int)
        c = response.json['containers']
        self.assertEqual(0, len(c))
        self.assertTrue(mock_container_create.called)

    @patch('zun.compute.api.API.container_create')
    @patch('zun.common.context.RequestContext.can')
    @patch('zun.network.neutron.NeutronAPI.get_neutron_network')
    @patch('zun.network.neutron.NeutronAPI.ensure_neutron_port_usable')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_with_public_network(
            self, mock_search, mock_ensure_port_usable, mock_get_network,
            mock_authorize, mock_container_create):
        fake_public_network = {'id': 'fakepubnetid',
                               'router:external': True,
                               'shared': False}
        mock_get_network.return_value = fake_public_network
        # Create a container with a command
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"},'
                  '"nets": [{"network": "testpublicnet"}]}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')
        fake_admin_authorize = True
        mock_authorize.return_value = fake_admin_authorize
        self.assertEqual(202, response.status_int)

        fake_not_admin_authorize = False
        mock_authorize.return_value = fake_not_admin_authorize
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json',
                             expect_errors=True)
        self.assertEqual(403, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(
            "It is not allowed to create an interface on external network %s" %
            fake_public_network['id'], response.json['errors'][0]['detail'])
        self.assertTrue(mock_container_create.assert_not_called)

    @patch('zun.compute.api.API.container_create')
    @patch('zun.common.context.RequestContext.can')
    @patch('zun.network.neutron.NeutronAPI.get_neutron_network')
    @patch('zun.network.neutron.NeutronAPI.ensure_neutron_port_usable')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_with_ip_addr(
            self, mock_search, mock_ensure_port_usable, mock_get_network,
            mock_authorize, mock_container_create):
        fake_network = {'id': 'fakenetid'}
        mock_get_network.return_value = fake_network
        # Create a container with a command
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"},'
                  '"nets": [{"network": "fakenetid", "v4-fixed-ip": '
                  '"10.0.0.10"}]}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')
        fake_admin_authorize = True
        mock_authorize.return_value = fake_admin_authorize
        self.assertEqual(202, response.status_int)

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_show')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.common.context.RequestContext.can')
    @patch('zun.volume.cinder_api.CinderAPI.create_volume')
    @patch('zun.volume.cinder_api.CinderAPI.ensure_volume_usable')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_with_create_new_volume(
            self, mock_search, mock_ensure_volume_usable, mock_create_volume,
            mock_authorize, mock_container_create, mock_container_show,
            mock_neutron_get_network):
        fake_network = {'id': 'foo'}
        mock_neutron_get_network.return_value = fake_network
        fake_volume_id = 'fakevolid'
        fake_volume = mock.Mock(id=fake_volume_id)
        mock_create_volume.return_value = fake_volume
        # Create a container with a command
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"mounts": [{"destination": "d", '
                  '"size": "5"}]}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(202, response.status_int)
        # get all containers
        container = objects.Container.list(self.context)[0]
        container.status = 'Creating'
        mock_container_show.return_value = container
        response = self.app.get('/v1/containers/')
        self.assertEqual(200, response.status_int)
        self.assertEqual(2, len(response.json))
        c = response.json['containers'][0]
        self.assertIsNotNone(c.get('uuid'))
        self.assertEqual('MyDocker', c.get('name'))
        self.assertEqual(["env"], c.get('command'))
        self.assertEqual('Creating', c.get('status'))
        self.assertEqual('512', c.get('memory'))
        self.assertIn('host', c)
        requested_networks = \
            mock_container_create.call_args[1]['requested_networks']
        self.assertEqual(1, len(requested_networks))
        self.assertEqual(fake_network['id'], requested_networks[0]['network'])
        mock_create_volume.assert_called_once()
        mock_ensure_volume_usable.assert_called_once()
        requested_volumes = \
            mock_container_create.call_args[1]['requested_volumes']
        self.assertEqual(1, len(requested_volumes))
        self.assertEqual(
            fake_volume_id,
            requested_volumes[container.uuid][0].cinder_volume_id)

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_show')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.common.context.RequestContext.can')
    @patch('zun.volume.cinder_api.CinderAPI.create_volume')
    @patch('zun.volume.cinder_api.CinderAPI.ensure_volume_usable')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_with_injected_file(
            self, mock_search, mock_ensure_volume_usable, mock_create_volume,
            mock_authorize, mock_container_create, mock_container_show,
            mock_neutron_get_network):
        fake_network = {'id': 'foo'}
        mock_neutron_get_network.return_value = fake_network
        # Create a container with a command
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"mounts": [{"destination": "d", "source": "hello",'
                  '            "type": "bind"}]}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(202, response.status_int)
        # get all containers
        container = objects.Container.list(self.context)[0]
        container.status = 'Creating'
        mock_container_show.return_value = container
        response = self.app.get('/v1/containers/')
        self.assertEqual(200, response.status_int)
        self.assertEqual(2, len(response.json))
        c = response.json['containers'][0]
        self.assertIsNotNone(c.get('uuid'))
        self.assertEqual('MyDocker', c.get('name'))
        self.assertEqual(["env"], c.get('command'))
        self.assertEqual('Creating', c.get('status'))
        self.assertEqual('512', c.get('memory'))
        self.assertIn('host', c)
        requested_networks = \
            mock_container_create.call_args[1]['requested_networks']
        self.assertEqual(1, len(requested_networks))
        self.assertEqual(fake_network['id'], requested_networks[0]['network'])
        mock_create_volume.assert_not_called()
        mock_ensure_volume_usable.assert_not_called()
        requested_volumes = \
            mock_container_create.call_args[1]['requested_volumes']
        self.assertEqual(1, len(requested_volumes))
        self.assertIsNone(
            requested_volumes[container.uuid][0].cinder_volume_id)
        self.assertEqual(
            'local', requested_volumes[container.uuid][0].volume_provider)

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_show')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_with_restart_policy_always_and_retrycount(
            self,
            mock_search,
            mock_container_create,
            mock_container_show,
            mock_neutron_get_network):
        mock_container_create.side_effect = lambda x, y, **z: y
        # Create a container with a command
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"restart_policy": {"Name": "always",'
                  '"MaximumRetryCount": "1"}}')
        with self.assertRaisesRegex(
                AppError, "maximum retry count not valid with"):
            self.post('/v1/containers/',
                      params=params,
                      content_type='application/json')
        self.assertTrue(mock_container_create.assert_not_called)
        mock_neutron_get_network.assert_called_once()

    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_invalid_long_name(self, mock_search,
                                                mock_container_create):
        # Long name
        params = ('{"name": "' + 'i' * 256 + '", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512"}')
        self.assertRaises(AppError, self.post, '/v1/containers/',
                          params=params, content_type='application/json')
        self.assertTrue(mock_container_create.assert_not_called)

    @patch('zun.objects.Container.list')
    def test_get_all_containers(self, mock_container_list):
        test_container = utils.get_test_container()
        containers = [objects.Container(self.context, **test_container)]
        mock_container_list.return_value = containers

        response = self.get('/v1/containers/')

        mock_container_list.assert_called_once_with(mock.ANY,
                                                    1000, None, 'id', 'asc',
                                                    filters={})
        context = mock_container_list.call_args[0][0]
        self.assertIs(False, context.all_projects)
        self.assertEqual(200, response.status_int)
        actual_containers = response.json['containers']
        self.assertEqual(1, len(actual_containers))
        self.assertEqual(test_container['uuid'],
                         actual_containers[0].get('uuid'))
        self.assertNotIn('host', actual_containers[0])

    @patch('zun.common.context.RequestContext.can')
    @patch('zun.objects.Container.list')
    def test_get_all_containers_by_admin(self, mock_container_list, mock_can):
        mock_can.return_value = True
        test_container = utils.get_test_container()
        containers = [objects.Container(self.context, **test_container)]
        mock_container_list.return_value = containers

        response = self.get('/v1/containers/')

        mock_container_list.assert_called_once_with(mock.ANY,
                                                    1000, None, 'id', 'asc',
                                                    filters={})
        context = mock_container_list.call_args[0][0]
        self.assertIs(False, context.all_projects)
        self.assertEqual(200, response.status_int)
        actual_containers = response.json['containers']
        self.assertEqual(1, len(actual_containers))
        self.assertEqual(test_container['uuid'],
                         actual_containers[0].get('uuid'))
        self.assertIn('host', actual_containers[0])

    @patch('zun.common.policy.enforce')
    @patch('zun.objects.Container.list')
    def test_get_all_containers_all_projects(self, mock_container_list,
                                             mock_policy):
        mock_policy.return_value = True
        test_container = utils.get_test_container()
        containers = [objects.Container(self.context, **test_container)]
        mock_container_list.return_value = containers

        response = self.get('/v1/containers/?all_projects=1')

        mock_container_list.assert_called_once_with(mock.ANY,
                                                    1000, None, 'id', 'asc',
                                                    filters={})
        context = mock_container_list.call_args[0][0]
        self.assertIs(True, context.all_projects)
        self.assertEqual(200, response.status_int)
        actual_containers = response.json['containers']
        self.assertEqual(1, len(actual_containers))
        self.assertEqual(test_container['uuid'],
                         actual_containers[0].get('uuid'))

    @patch('zun.objects.Container.list')
    def test_get_all_has_status_reason(
            self, mock_container_list):
        test_container = utils.get_test_container()
        containers = [objects.Container(self.context, **test_container)]
        mock_container_list.return_value = containers

        response = self.get('/v1/containers/')
        self.assertEqual(200, response.status_int)
        actual_containers = response.json['containers']
        self.assertEqual(1, len(actual_containers))
        self.assertEqual(test_container['uuid'],
                         actual_containers[0].get('uuid'))
        self.assertIn('status_reason', actual_containers[0].keys())

    @patch('zun.objects.Container.list')
    def test_get_all_containers_with_pagination_marker(self,
                                                       mock_container_list):
        container_list = []
        for id_ in range(4):
            test_container = utils.create_test_container(
                id=id_, uuid=uuidutils.generate_uuid(),
                name='container' + str(id_), context=self.context)
            container_list.append(objects.Container(self.context,
                                                    **test_container))
        mock_container_list.return_value = container_list[-1:]
        response = self.get('/v1/containers/?limit=3&marker=%s'
                            % container_list[2].uuid)

        self.assertEqual(200, response.status_int)
        actual_containers = response.json['containers']
        self.assertEqual(1, len(actual_containers))
        self.assertEqual(container_list[-1].uuid,
                         actual_containers[0].get('uuid'))

    @patch('zun.objects.Container.list')
    def test_get_all_containers_with_filter(self, mock_container_list):
        test_container = utils.get_test_container()
        containers = [objects.Container(self.context, **test_container)]
        mock_container_list.return_value = containers

        response = self.get('/v1/containers/?name=fake-name')

        mock_container_list.assert_called_once_with(
            mock.ANY, 1000, None, 'id', 'asc', filters={'name': 'fake-name'})
        self.assertEqual(200, response.status_int)
        actual_containers = response.json['containers']
        self.assertEqual(1, len(actual_containers))
        self.assertEqual(test_container['uuid'],
                         actual_containers[0].get('uuid'))

    @patch('zun.objects.Container.list')
    def test_get_all_containers_with_filter_disallow(
            self, mock_container_list):
        test_container = utils.get_test_container()
        containers = [objects.Container(self.context, **test_container)]
        mock_container_list.return_value = containers

        response = self.get('/v1/containers/?host=fake-name',
                            expect_errors=True)
        self.assertEqual(403, response.status_int)
        self.assertEqual('application/json', response.content_type)
        rule = 'container:get_one:host'
        self.assertEqual(
            "Policy doesn't allow %s to be performed." % rule,
            response.json['errors'][0]['detail'])

    @patch('zun.objects.Container.list')
    def test_get_all_containers_with_unknown_parameter(
            self, mock_container_list):
        test_container = utils.get_test_container()
        containers = [objects.Container(self.context, **test_container)]
        mock_container_list.return_value = containers

        response = self.get('/v1/containers/?unknown=fake-name',
                            expect_errors=True)

        mock_container_list.assert_not_called()
        self.assertEqual(400, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual("Unknown parameters: unknown",
                         response.json['errors'][0]['detail'])

    @patch('zun.objects.Container.list')
    def test_get_all_containers_with_exception(self, mock_container_list):
        test_container = utils.get_test_container()
        containers = [objects.Container(self.context, **test_container)]
        mock_container_list.return_value = containers
        response = self.get('/v1/containers/')
        mock_container_list.assert_called_once_with(mock.ANY,
                                                    1000, None, 'id', 'asc',
                                                    filters={})
        self.assertEqual(200, response.status_int)
        actual_containers = response.json['containers']
        self.assertEqual(1, len(actual_containers))
        self.assertEqual(test_container['uuid'],
                         actual_containers[0].get('uuid'))

    @patch('zun.compute.api.API.container_show')
    @patch('zun.objects.Container.get_by_uuid')
    def test_get_one_by_uuid(self, mock_container_get_by_uuid,
                             mock_container_show):
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_container_get_by_uuid.return_value = test_container_obj
        mock_container_show.return_value = test_container_obj

        response = self.get('/v1/containers/%s/' % test_container['uuid'])

        mock_container_get_by_uuid.assert_called_once_with(
            mock.ANY,
            test_container['uuid'])
        context = mock_container_get_by_uuid.call_args[0][0]
        self.assertIs(False, context.all_projects)
        self.assertEqual(200, response.status_int)
        self.assertEqual(test_container['uuid'],
                         response.json['uuid'])
        self.assertNotIn('host', response.json)

    @patch('zun.common.context.RequestContext.can')
    @patch('zun.compute.api.API.container_show')
    @patch('zun.objects.Container.get_by_uuid')
    def test_get_one_by_admin(self, mock_container_get_by_uuid,
                              mock_container_show, mock_can):
        mock_can.return_value = True
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_container_get_by_uuid.return_value = test_container_obj
        mock_container_show.return_value = test_container_obj

        response = self.get('/v1/containers/%s/' % test_container['uuid'])

        mock_container_get_by_uuid.assert_called_once_with(
            mock.ANY,
            test_container['uuid'])
        context = mock_container_get_by_uuid.call_args[0][0]
        self.assertIs(False, context.all_projects)
        self.assertEqual(200, response.status_int)
        self.assertEqual(test_container['uuid'],
                         response.json['uuid'])
        self.assertIn('host', response.json)

    @patch('zun.common.policy.enforce')
    @patch('zun.compute.api.API.container_show')
    @patch('zun.objects.Container.get_by_uuid')
    def test_get_one_by_uuid_all_projects(self, mock_container_get_by_uuid,
                                          mock_container_show, mock_policy):
        mock_policy.return_value = True
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_container_get_by_uuid.return_value = test_container_obj
        mock_container_show.return_value = test_container_obj

        response = self.get('/v1/containers/%s/?all_projects=1' %
                            test_container['uuid'])

        mock_container_get_by_uuid.assert_called_once_with(
            mock.ANY,
            test_container['uuid'])
        context = mock_container_get_by_uuid.call_args[0][0]
        self.assertIs(True, context.all_projects)
        self.assertEqual(200, response.status_int)
        self.assertEqual(test_container['uuid'],
                         response.json['uuid'])

    @patch('zun.objects.ComputeNode.get_by_name')
    @patch('zun.compute.api.API.container_update')
    @patch('zun.objects.Container.get_by_uuid')
    def test_patch_by_uuid(self, mock_container_get_by_uuid, mock_update,
                           mock_computenode):
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_container_get_by_uuid.return_value = test_container_obj
        mock_update.return_value = test_container_obj
        test_host = utils.get_test_compute_node()
        numa = objects.numa.NUMATopology._from_dict(test_host['numa_topology'])
        test_host['numa_topology'] = numa
        test_host_obj = objects.ComputeNode(self.context, **test_host)
        mock_computenode.return_value = test_host_obj
        params = {'cpu': 1}
        container_uuid = test_container.get('uuid')
        response = self.patch_json(
            '/containers/%s/' % container_uuid,
            params=params)

        self.assertEqual(200, response.status_int)
        self.assertTrue(mock_update.called)
        self.assertNotIn('host', response.json)

    @patch('zun.common.context.RequestContext.can')
    @patch('zun.objects.ComputeNode.get_by_name')
    @patch('zun.compute.api.API.container_update')
    @patch('zun.objects.Container.get_by_uuid')
    def test_patch_by_admin(self, mock_container_get_by_uuid, mock_update,
                            mock_computenode, mock_can):
        mock_can.return_value = True
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_container_get_by_uuid.return_value = test_container_obj
        mock_update.return_value = test_container_obj
        test_host = utils.get_test_compute_node()
        numa = objects.numa.NUMATopology._from_dict(test_host['numa_topology'])
        test_host['numa_topology'] = numa
        test_host_obj = objects.ComputeNode(self.context, **test_host)
        mock_computenode.return_value = test_host_obj
        params = {'cpu': 1}
        container_uuid = test_container.get('uuid')
        response = self.patch_json(
            '/containers/%s/' % container_uuid,
            params=params)

        self.assertEqual(200, response.status_int)
        self.assertTrue(mock_update.called)
        self.assertIn('host', response.json)

    def _action_test(self, test_container_obj, action, ident_field,
                     mock_container_action, status_code, query_param=''):
        ident = test_container_obj.uuid
        get_by_ident_loc = 'zun.objects.Container.get_by_%s' % ident_field
        with patch(get_by_ident_loc) as mock_get_by_indent:
            mock_get_by_indent.return_value = test_container_obj
            response = self.post('/v1/containers/%s/%s/?%s' %
                                 (ident, action, query_param))
            self.assertEqual(status_code, response.status_int)

            # Only PUT should work, others like GET should fail
            self.assertRaises(AppError, self.get,
                              ('/v1/containers/%s/%s/' %
                               (ident, action)))
        if query_param:
            value = query_param.split('=')[1]
            mock_container_action.assert_called_once_with(
                mock.ANY, test_container_obj, value)
        else:
            mock_container_action.assert_called_once_with(
                mock.ANY, test_container_obj)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_start')
    def test_start_by_uuid(self, mock_container_start, mock_validate):
        test_container_obj = objects.Container(self.context,
                                               **utils.get_test_container())
        self._action_test(test_container_obj, 'start', 'uuid',
                          mock_container_start, 202)

    def test_start_by_uuid_invalid_state(self):
        uuid = uuidutils.generate_uuid()
        test_object = utils.create_test_container(context=self.context,
                                                  uuid=uuid, status='Running')
        with self.assertRaisesRegex(
                AppError, "Cannot start container %s in Running state" % uuid):
            self.post('/v1/containers/%s/%s/' % (test_object.uuid,
                                                 'start'))

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_stop')
    def test_stop_by_uuid(self, mock_container_stop, mock_validate):
        test_container_obj = objects.Container(self.context,
                                               **utils.get_test_container())
        self._action_test(test_container_obj, 'stop', 'uuid',
                          mock_container_stop, 202,
                          query_param='timeout=10')

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_stop')
    def test_stop_by_name_invalid_timeout_value(self,
                                                mock_container_stop,
                                                mock_validate):
        test_container_obj = objects.Container(self.context,
                                               **utils.get_test_container())
        with self.assertRaisesRegex(AppError,
                                    "Invalid input for query parameters"):
            self._action_test(test_container_obj, 'stop', 'name',
                              mock_container_stop, 202,
                              query_param='timeout=xyz')

    def test_stop_by_uuid_invalid_state(self):
        uuid = uuidutils.generate_uuid()
        test_object = utils.create_test_container(context=self.context,
                                                  uuid=uuid, status='Stopped')
        with self.assertRaisesRegex(
                AppError, "Cannot stop container %s in Stopped state" % uuid):
            self.post('/v1/containers/%s/%s/' % (test_object.uuid,
                                                 'stop'))

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_pause')
    def test_pause_by_uuid(self, mock_container_pause, mock_validate):
        test_container_obj = objects.Container(self.context,
                                               **utils.get_test_container())
        self._action_test(test_container_obj, 'pause', 'uuid',
                          mock_container_pause, 202)

    def test_pause_by_uuid_invalid_state(self):
        uuid = uuidutils.generate_uuid()
        test_object = utils.create_test_container(context=self.context,
                                                  uuid=uuid, status='Stopped')
        with self.assertRaisesRegex(
                AppError, "Cannot pause container %s in Stopped state" % uuid):
            self.post('/v1/containers/%s/%s/' % (test_object.uuid,
                                                 'pause'))

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_unpause')
    def test_unpause_by_uuid(self, mock_container_unpause, mock_validate):
        test_container_obj = objects.Container(self.context,
                                               **utils.get_test_container())
        self._action_test(test_container_obj, 'unpause', 'uuid',
                          mock_container_unpause, 202)

    def test_unpause_by_uuid_invalid_state(self):
        uuid = uuidutils.generate_uuid()
        test_object = utils.create_test_container(context=self.context,
                                                  uuid=uuid, status='Running')
        with self.assertRaisesRegex(
                AppError,
                "Cannot unpause container %s in Running state" % uuid):
            self.post('/v1/containers/%s/%s/' % (test_object.uuid,
                                                 'unpause'))

    @patch('zun.compute.api.API.container_rebuild')
    @patch('zun.common.policy.enforce')
    def test_rebuild_container(self, mock_policy, mock_rebuild):
        mock_policy.return_value = True
        uuid = uuidutils.generate_uuid()
        utils.create_test_container(context=self.context,
                                    uuid=uuid, name="container")
        url = '/v1/containers/%s/rebuild' % uuid
        response = self.post(url)
        self.assertEqual(202, response.status_int)
        mock_rebuild.assert_called_once()

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_reboot')
    def test_reboot_by_uuid(self, mock_container_reboot, mock_validate):
        test_container_obj = objects.Container(self.context,
                                               **utils.get_test_container())
        with patch.object(test_container_obj, 'save') as mock_save:
            self._action_test(test_container_obj, 'reboot', 'uuid',
                              mock_container_reboot, 202,
                              query_param='timeout=10')
            mock_save.assert_called_once()

    def test_reboot_by_uuid_invalid_state(self):
        uuid = uuidutils.generate_uuid()
        test_object = utils.create_test_container(context=self.context,
                                                  uuid=uuid, status='Paused')
        with self.assertRaisesRegex(
                AppError, "Cannot reboot container %s in Paused state" % uuid):
            self.post('/v1/containers/%s/%s/' % (test_object.uuid,
                                                 'reboot'))

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_reboot')
    def test_reboot_by_name_wrong_timeout_value(self, mock_container_reboot,
                                                mock_validate):
        test_container_obj = objects.Container(self.context,
                                               **utils.get_test_container())
        with self.assertRaisesRegex(AppError,
                                    "Invalid input for query parameters"):
            self._action_test(test_container_obj, 'reboot', 'name',
                              mock_container_reboot, 202,
                              query_param='timeout=xyz')

    @patch('zun.compute.api.API.container_logs')
    @patch('zun.objects.Container.get_by_uuid')
    def test_get_logs_by_uuid(self, mock_get_by_uuid, mock_container_logs):
        mock_container_logs.return_value = "test"
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_uuid = test_container.get('uuid')
        response = self.get('/v1/containers/%s/logs/' % container_uuid)

        self.assertEqual(200, response.status_int)
        mock_container_logs.assert_called_once_with(
            mock.ANY, test_container_obj, True, True, False, 'all', None)

    @patch('zun.compute.api.API.container_logs')
    @patch('zun.objects.Container.get_by_uuid')
    def test_get_logs_with_options_by_uuid(self, mock_get_by_uuid,
                                           mock_container_logs):
        mock_container_logs.return_value = "test"
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_uuid = test_container.get('uuid')
        response = self.get(
            '/v1/containers/%s/logs?stderr=True&stdout=True'
            '&timestamps=False&tail=1&since=100000000' % container_uuid)
        self.assertEqual(200, response.status_int)
        mock_container_logs.assert_called_once_with(
            mock.ANY, test_container_obj, True, True, False, '1', '100000000')

    @patch('zun.compute.api.API.container_logs')
    @patch('zun.objects.Container.get_by_uuid')
    def test_get_logs_put_fails(self, mock_get_by_uuid, mock_container_logs):
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_uuid = test_container.get('uuid')
        self.assertRaises(AppError, self.post,
                          '/v1/containers/%s/logs/' % container_uuid)
        self.assertFalse(mock_container_logs.called)

    @patch('zun.compute.api.API.container_logs')
    @patch('zun.objects.Container.get_by_uuid')
    def test_get_logs_with_invalid_since(self, mock_get_by_uuid,
                                         mock_container_logs):
        invalid_sinces = ['x11', '11x', '2000-01-01 01:01:01']
        for value in invalid_sinces:
            test_container = utils.get_test_container()
            test_container_obj = objects.Container(self.context,
                                                   **test_container)
            mock_get_by_uuid.return_value = test_container_obj

            container_uuid = test_container.get('uuid')
            params = {'since': value}

            self.assertRaises(AppError, self.post,
                              '/v1/containers/%s/logs' %
                              container_uuid, params)
            self.assertFalse(mock_container_logs.called)

    def test_get_logs_with_invalid_state(self):
        uuid = uuidutils.generate_uuid()
        test_object = utils.create_test_container(context=self.context,
                                                  uuid=uuid, status='Creating')
        with self.assertRaisesRegex(
                AppError,
                "Cannot logs container %s in Creating state" % uuid):
            self.get('/v1/containers/%s/logs/' % test_object.uuid)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_exec')
    @patch('zun.objects.Container.get_by_uuid')
    def test_execute_command_by_uuid(self, mock_get_by_uuid,
                                     mock_container_exec, mock_validate):
        mock_container_exec.return_value = ""
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_uuid = test_container.get('uuid')
        url = '/v1/containers/%s/%s/' % (container_uuid, 'execute')
        cmd = {'command': 'ls'}
        response = self.post(url, cmd)
        self.assertEqual(200, response.status_int)
        mock_container_exec.assert_called_once_with(
            mock.ANY, test_container_obj, cmd['command'], True, False)

    def test_exec_command_by_uuid_invalid_state(self):
        uuid = uuidutils.generate_uuid()
        test_object = utils.create_test_container(context=self.context,
                                                  uuid=uuid, status='Stopped')
        cmd = {'command': 'ls'}
        with self.assertRaisesRegex(
                AppError,
                "Cannot execute container %s in Stopped state" % uuid):
            self.post('/v1/containers/%s/%s/' % (test_object.uuid,
                                                 'execute'), cmd)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_exec')
    @patch('zun.objects.Container.get_by_uuid')
    def test_execute_without_command_by_uuid(self, mock_get_by_uuid,
                                             mock_container_exec,
                                             mock_validate):
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_uuid = test_container.get('uuid')
        cmd = {'command': ''}
        self.assertRaises(AppError, self.post,
                          '/v1/containers/%s/execute' %
                          container_uuid, cmd)
        self.assertFalse(mock_container_exec.called)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_delete')
    @patch('zun.objects.Container.get_by_uuid')
    def test_delete_container_by_uuid(self, mock_get_by_uuid,
                                      mock_container_delete, mock_validate):
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_uuid = test_container.get('uuid')
        response = self.delete('/v1/containers/%s/' % container_uuid)

        self.assertEqual(204, response.status_int)
        mock_container_delete.assert_called_once_with(
            mock.ANY, test_container_obj, False)
        context = mock_container_delete.call_args[0][0]
        self.assertIs(False, context.all_projects)

    @patch('zun.common.policy.enforce')
    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_delete')
    @patch('zun.objects.Container.get_by_uuid')
    def test_delete_container_by_uuid_all_projects(self, mock_get_by_uuid,
                                                   mock_container_delete,
                                                   mock_validate, mock_policy):
        mock_policy.return_value = True
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_uuid = test_container.get('uuid')
        response = self.delete('/v1/containers/%s/?all_projects=1' %
                               container_uuid)

        self.assertEqual(204, response.status_int)
        mock_container_delete.assert_called_once_with(
            mock.ANY, test_container_obj, False)
        context = mock_container_delete.call_args[0][0]
        self.assertIs(True, context.all_projects)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_stop')
    @patch('zun.compute.api.API.container_delete')
    @patch('zun.objects.Container.get_by_uuid')
    def test_delete_container_by_uuid_with_stop(self, mock_get_by_uuid,
                                                mock_container_stop,
                                                mock_container_delete,
                                                mock_validate):
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_uuid = test_container.get('uuid')
        response = self.delete('/v1/containers/%s?stop=True' %
                               container_uuid)

        self.assertEqual(204, response.status_int)

    def test_delete_by_uuid_invalid_state(self):
        uuid = uuidutils.generate_uuid()
        test_object = utils.create_test_container(context=self.context,
                                                  uuid=uuid, status='Running')
        with self.assertRaisesRegex(
                AppError,
                "Cannot delete container %s in Running state" % uuid):
            self.delete('/v1/containers/%s' % (test_object.uuid))

    @patch('zun.common.policy.enforce')
    def test_delete_force_by_uuid_invalid_state(self, mock_policy):
        mock_policy.return_value = True
        uuid = uuidutils.generate_uuid()
        test_object = utils.create_test_container(context=self.context,
                                                  uuid=uuid, status='Paused')
        with self.assertRaisesRegex(
                AppError,
                "Cannot delete_force container %s in Paused state" % uuid):
            self.delete('/v1/containers/%s?force=True' % test_object.uuid)

    @patch('zun.common.policy.enforce')
    @patch('zun.compute.api.API.container_delete')
    def test_delete_by_uuid_invalid_state_force_true(self, mock_delete,
                                                     mock_policy):
        mock_policy.return_value = True
        uuid = uuidutils.generate_uuid()
        test_object = utils.create_test_container(context=self.context,
                                                  uuid=uuid, status='Running')
        response = self.delete('/v1/containers/%s?force=True' % (
            test_object.uuid))
        self.assertEqual(204, response.status_int)

    @patch('zun.compute.api.API.container_delete')
    def test_delete_by_uuid_with_force_wrong(self, mock_delete):
        uuid = uuidutils.generate_uuid()
        test_object = utils.create_test_container(context=self.context,
                                                  uuid=uuid)
        mock_delete.side_effect = exception.InvalidValue
        self.assertRaises(AppError, self.delete,
                          '/v1/containers/%s?force=wrong' % test_object.uuid)
        self.assertTrue(mock_delete.assert_not_called)

    def test_delete_container_with_uuid_not_found(self):
        uuid = uuidutils.generate_uuid()
        self.assertRaises(AppError, self.delete,
                          '/v1/containers/%s' % uuid)

    @patch('zun.objects.Container.destroy')
    @patch('zun.common.utils.validate_container_state')
    @patch('zun.objects.Container.get_by_uuid')
    def test_delete_container_without_host(self, mock_get_by_uuid,
                                           mock_validate, mock_destroy):
        test_container = utils.get_test_container(host="")
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj
        container_uuid = test_container.get('uuid')
        response = self.delete('/v1/containers/%s/' % container_uuid)
        self.assertEqual(204, response.status_int)
        mock_validate.assert_called_once()
        mock_destroy.assert_called_once()

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_kill')
    @patch('zun.objects.Container.get_by_uuid')
    def test_kill_container_by_uuid(self,
                                    mock_get_by_uuid, mock_container_kill,
                                    mock_validate):
        test_container_obj = objects.Container(self.context,
                                               **utils.get_test_container())
        mock_container_kill.return_value = test_container_obj
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_uuid = test_container.get('uuid')
        url = '/v1/containers/%s/%s/' % (container_uuid, 'kill')
        cmd = {'signal': '9'}
        response = self.post(url, cmd)
        self.assertEqual(202, response.status_int)
        mock_container_kill.assert_called_once_with(
            mock.ANY, test_container_obj, cmd['signal'])

    def test_kill_by_uuid_invalid_state(self):
        uuid = uuidutils.generate_uuid()
        test_object = utils.create_test_container(context=self.context,
                                                  uuid=uuid, status='Stopped')
        body = {'signal': 9}
        with self.assertRaisesRegex(
                AppError, "Cannot kill container %s in Stopped state" % uuid):
            self.post('/v1/containers/%s/%s/' % (test_object.uuid,
                                                 'kill'), body)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_kill')
    @patch('zun.objects.Container.get_by_uuid')
    def test_kill_container_which_not_exist(self,
                                            mock_get_by_uuid,
                                            mock_container_kill,
                                            mock_validate):
        mock_container_kill.return_value = ""
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj
        mock_container_kill.side_effect = Exception

        container_uuid = "edfe2a25-2901-438d-8157-fffffd68d051"
        self.assertRaises(AppError, self.post,
                          '/v1/containers/%s/%s/' % (container_uuid, 'kill'))
        self.assertTrue(mock_container_kill.called)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_kill')
    @patch('zun.objects.Container.get_by_uuid')
    def test_kill_container_with_exception(self,
                                           mock_get_by_uuid,
                                           mock_container_kill,
                                           mock_validate):
        mock_container_kill.return_value = ""
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj
        mock_container_kill.side_effect = Exception

        container_uuid = test_container.get('uuid')
        self.assertRaises(AppError, self.post,
                          '/v1/containers/%s/%s/' % (container_uuid, 'kill'))
        self.assertTrue(mock_container_kill.called)

    @patch('zun.compute.api.API.container_kill')
    @patch('zun.objects.Container.get_by_uuid')
    def test_kill_container_with_invalid_singal(self,
                                                mock_get_by_uuid,
                                                mock_container_kill):
        invalid_signal = ['11x', 'x11']
        for value in invalid_signal:
            test_container = utils.get_test_container()
            test_container_obj = objects.Container(self.context,
                                                   **test_container)
            mock_get_by_uuid.return_value = test_container_obj

            container_uuid = test_container.get('uuid')
            params = {'signal': value}
            with self.assertRaisesRegex(
                    AppError, "Bad response: 400 Bad Request"):
                self.post('/v1/containers/%s/kill/' %
                          container_uuid, params)
            self.assertFalse(mock_container_kill.called)

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.container_create')
    @patch('zun.compute.api.API.image_search')
    def test_create_container_resp_has_image_driver(self, mock_search,
                                                    mock_container_create,
                                                    mock_neutron_get_network):
        mock_container_create.side_effect = lambda x, y, **z: y
        # Create a container with a command
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512",'
                  '"environment": {"key1": "val1", "key2": "val2"},'
                  '"image_driver": "glance"}')
        response = self.post('/v1/containers/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(202, response.status_int)
        self.assertIn('image_driver', response.json.keys())
        self.assertEqual('glance', response.json.get('image_driver'))

    @patch('zun.compute.api.API.container_attach')
    @patch('zun.objects.Container.get_by_uuid')
    def test_attach_container_by_uuid(self, mock_get_by_uuid,
                                      mock_container_attach):
        mock_container_attach.return_value = "ws://test"
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context,
                                               **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_uuid = test_container.get('uuid')
        response = self.get('/v1/containers/%s/attach/' % container_uuid)

        self.assertEqual(200, response.status_int)
        mock_container_attach.assert_called_once_with(
            mock.ANY, test_container_obj)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_attach')
    @patch('zun.objects.Container.get_by_uuid')
    def test_attach_container_with_exception(self,
                                             mock_get_by_uuid,
                                             mock_container_attach,
                                             mock_validate):
        mock_container_attach.return_value = ""
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj
        mock_container_attach.side_effect = Exception

        container_uuid = test_container.get('uuid')
        self.assertRaises(AppError, self.get,
                          '/v1/containers/%s/attach/' % container_uuid)
        self.assertTrue(mock_container_attach.called)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_resize')
    @patch('zun.objects.Container.get_by_name')
    def test_resize_by_uuid(self, mock_get_by_uuid, mock_container_resize,
                            mock_validate):
        test_container_obj = objects.Container(self.context,
                                               **utils.get_test_container())
        mock_container_resize.return_value = test_container_obj
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_name = test_container.get('name')
        url = '/v1/containers/%s/%s/' % (container_name, 'resize')
        cmd = {'h': '100', 'w': '100'}
        response = self.post(url, cmd)
        self.assertEqual(200, response.status_int)
        mock_container_resize.assert_called_once_with(
            mock.ANY, test_container_obj, cmd['h'], cmd['w'])

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_resize')
    @patch('zun.objects.Container.get_by_uuid')
    def test_resize_with_exception(self, mock_get_by_uuid,
                                   mock_container_resize, mock_validate):
        mock_container_resize.return_value = ""
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj
        mock_container_resize.side_effect = Exception

        container_uuid = test_container.get('uuid')
        body = {'h': '100', 'w': '100'}
        self.assertRaises(AppError, self.post,
                          '/v1/containers/%s/%s/' %
                          (container_uuid, 'resize'), body)
        self.assertTrue(mock_container_resize.called)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.resize_container')
    @patch('zun.objects.Container.get_by_name')
    def test_resize_container(self, mock_get_by_uuid,
                              mock_resize_container, mock_validate):
        test_container_obj = objects.Container(self.context,
                                               **utils.get_test_container())
        mock_resize_container.return_value = test_container_obj
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_name = test_container.get('name')
        url = '/v1/containers/%s/resize_container/' % container_name
        params = {'cpu': 1, 'memory': '512'}
        response = self.post(url, params)
        self.assertEqual(202, response.status_int)
        mock_resize_container.assert_called_once_with(
            mock.ANY, test_container_obj, params)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_top')
    @patch('zun.objects.Container.get_by_uuid')
    def test_top_command_by_uuid(self, mock_get_by_uuid,
                                 mock_container_top, mock_validate):
        mock_container_top.return_value = ""
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_uuid = test_container.get('uuid')
        response = self.get('/v1/containers/%s/top?ps_args=aux' %
                            container_uuid)
        self.assertEqual(200, response.status_int)
        self.assertTrue(mock_container_top.called)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_top')
    @patch('zun.objects.Container.get_by_uuid')
    def test_top_command_invalid_ps(self, mock_get_by_uuid,
                                    mock_container_top, mock_validate):
        mock_container_top.return_value = ""
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj
        mock_container_top.side_effect = Exception

        container_uuid = test_container.get('uuid')
        self.assertRaises(AppError, self.get,
                          '/v1/containers/%s/top?ps_args=kkkk' %
                          container_uuid)
        self.assertTrue(mock_container_top.called)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_get_archive')
    @patch('zun.objects.Container.get_by_uuid')
    def test_get_archive_by_uuid(self,
                                 mock_get_by_uuid,
                                 container_get_archive,
                                 mock_validate):
        container_get_archive.return_value = ("", "")
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_uuid = test_container.get('uuid')
        url = '/v1/containers/%s/%s/' % (container_uuid, 'get_archive')
        cmd = {'path': '/home/1.txt'}
        response = self.get(url, cmd)
        self.assertEqual(200, response.status_int)
        container_get_archive.assert_called_once_with(
            mock.ANY, test_container_obj, cmd['path'], mock.ANY)

    def test_get_archive_by_uuid_invalid_state(self):
        uuid = uuidutils.generate_uuid()
        test_object = utils.create_test_container(context=self.context,
                                                  uuid=uuid, status='Error')
        with self.assertRaisesRegex(
                AppError,
                "Cannot get_archive container %s in Error state" % uuid):
            self.get('/v1/containers/%s/%s/' % (test_object.uuid,
                                                'get_archive'))

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_put_archive')
    @patch('zun.objects.Container.get_by_uuid')
    def test_put_archive_by_uuid(self,
                                 mock_get_by_uuid,
                                 container_put_archive,
                                 mock_validate):
        container_put_archive.return_value = ""
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_uuid = test_container.get('uuid')
        url = '/v1/containers/%s/%s/' % (container_uuid, 'put_archive')
        cmd = {'path': '/home/',
               'data': '/home/1.tar'}
        response = self.post(url, cmd)
        self.assertEqual(200, response.status_int)
        container_put_archive.assert_called_once_with(
            mock.ANY, test_container_obj, cmd['path'], cmd['data'], mock.ANY)

    def test_put_archive_by_uuid_invalid_state(self):
        uuid = uuidutils.generate_uuid()
        test_object = utils.create_test_container(context=self.context,
                                                  uuid=uuid, status='Error')
        with self.assertRaisesRegex(
                AppError,
                "Cannot put_archive container %s in Error state" % uuid):
            self.post('/v1/containers/%s/%s/' % (test_object.uuid,
                                                 'put_archive'))

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_stats')
    @patch('zun.objects.Container.get_by_uuid')
    def test_stats_container_by_uuid(self, mock_get_by_uuid,
                                     mock_container_stats, mock_validate):
        mock_container_stats.return_value = ""
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj

        container_uuid = test_container.get('uuid')
        url = '/v1/containers/%s/stats'\
              % container_uuid
        response = self.get(url)
        self.assertEqual(200, response.status_int)
        self.assertTrue(mock_container_stats.called)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_commit')
    @patch('zun.objects.Container.get_by_name')
    def test_commit_by_name(self, mock_get_by_name,
                            mock_container_commit, mock_validate):

        test_container_obj = objects.Container(self.context,
                                               **utils.get_test_container())
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_name.return_value = test_container_obj
        mock_container_commit.return_value = None
        container_name = test_container.get('name')
        url = '/v1/containers/%s/%s/' % (container_name, 'commit')
        cmd = {'repository': 'repo', 'tag': 'tag'}
        response = self.post(url, cmd)
        self.assertEqual(202, response.status_int)
        mock_container_commit.assert_called_once_with(
            mock.ANY, test_container_obj, cmd['repository'], cmd['tag'])

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_commit')
    @patch('zun.objects.Container.get_by_uuid')
    def test_commit_by_uuid(self, mock_get_by_uuid,
                            mock_container_commit, mock_validate):

        test_container_obj = objects.Container(self.context,
                                               **utils.get_test_container())
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_by_uuid.return_value = test_container_obj
        mock_container_commit.return_value = None
        container_uuid = test_container.get('uuid')
        url = '/v1/containers/%s/%s/' % (container_uuid, 'commit')
        cmd = {'repository': 'repo', 'tag': 'tag'}
        response = self.post(url, cmd)
        self.assertEqual(202, response.status_int)
        mock_container_commit.assert_called_once_with(
            mock.ANY, test_container_obj, cmd['repository'], cmd['tag'])

    def test_commit_by_uuid_invalid_state(self):
        uuid = uuidutils.generate_uuid()
        cmd = {'repository': 'repo', 'tag': 'tag'}
        utils.create_test_container(context=self.context,
                                    uuid=uuid, status='Error')
        with self.assertRaisesRegex(
                AppError, "Cannot commit container %s in Error state" % uuid):
            self.post('/v1/containers/%s/commit/' % uuid, cmd)

    @patch('zun.common.utils.validate_container_state')
    @patch('zun.compute.api.API.container_exec_resize')
    @patch('zun.api.utils.get_resource')
    def test_execute_resize_container_exec(
            self, mock_get_resource, mock_exec_resize, mock_validate):
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_resource.return_value = test_container_obj
        mock_exec_resize.return_value = None
        container_name = test_container.get('name')
        url = '/v1/containers/%s/%s/' % (container_name, 'execute_resize')
        fake_exec_id = ('7df36611fa1fc855618c2c643835d41d'
                        'ac3fe568e7688f0bae66f7bcb3cccc6c')
        kwargs = {'exec_id': fake_exec_id, 'h': '100', 'w': '100'}
        response = self.post(url, kwargs)
        self.assertEqual(200, response.status_int)
        mock_exec_resize.assert_called_once_with(
            mock.ANY, test_container_obj, fake_exec_id, kwargs['h'],
            kwargs['w'])

    @mock.patch('zun.compute.api.API.add_security_group')
    @mock.patch('zun.network.neutron.NeutronAPI.find_resourceid_by_name_or_id')
    @mock.patch('zun.api.utils.get_resource')
    def test_add_security_group_by_uuid(self, mock_get_resource,
                                        mock_find_resourceid,
                                        mock_add_security_group):
        headers = {"OpenStack-API-Version": "container 1.14"}
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_resource.return_value = test_container_obj
        mock_find_resourceid.return_value = 'fake_security_group_id'
        container_name = test_container.get('name')
        security_group_id_to_add = '5f7cf831-9a9c-4e2b-87b2-6081667f852b'
        url = '/v1/containers/%s/%s?name=%s' % (container_name,
                                                'add_security_group',
                                                security_group_id_to_add)
        response = self.post(url, headers=headers)
        self.assertEqual(202, response.status_int)
        self.assertEqual('application/json', response.content_type)
        mock_find_resourceid.assert_called_once_with(
            'security_group', security_group_id_to_add, mock.ANY)
        mock_add_security_group.assert_called_once_with(
            mock.ANY, test_container_obj, 'fake_security_group_id')

    @mock.patch('zun.compute.api.API.add_security_group')
    @mock.patch('zun.network.neutron.NeutronAPI.find_resourceid_by_name_or_id')
    @mock.patch('zun.api.utils.get_resource')
    def test_add_security_group_not_found(self, mock_get_resource,
                                          mock_find_resourceid,
                                          mock_add_security_group):
        headers = {"OpenStack-API-Version": "container 1.14"}
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_resource.return_value = test_container_obj
        mock_find_resourceid.side_effect = n_exc.NotFound()
        container_name = test_container.get('name')
        security_group_to_add = '5f7cf831-9a9c-4e2b-87b2-6081667f852b'
        url = '/v1/containers/%s/%s?name=%s' % (container_name,
                                                'add_security_group',
                                                security_group_to_add)
        response = self.post(url, expect_errors=True, headers=headers)
        self.assertEqual(400, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(
            "Security group %s not found." % security_group_to_add,
            response.json['errors'][0]['detail'])

    @mock.patch('zun.compute.api.API.add_security_group')
    @mock.patch('zun.network.neutron.NeutronAPI.find_resourceid_by_name_or_id')
    @mock.patch('zun.api.utils.get_resource')
    def test_add_security_group_not_unique_match(self, mock_get_resource,
                                                 mock_find_resourceid,
                                                 mock_add_security_group):
        headers = {"OpenStack-API-Version": "container 1.14"}
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_resource.return_value = test_container_obj
        mock_find_resourceid.side_effect = n_exc.NeutronClientNoUniqueMatch()
        container_name = test_container.get('name')
        security_group_to_add = '5f7cf831-9a9c-4e2b-87b2-6081667f852b'
        url = '/v1/containers/%s/%s?name=%s' % (container_name,
                                                'add_security_group',
                                                security_group_to_add)
        response = self.post(url, expect_errors=True, headers=headers)
        self.assertEqual(409, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(
            "Multiple security group matches found for name %s, "
            "use an ID to be more specific." % security_group_to_add,
            response.json['errors'][0]['detail'])

    @patch('zun.network.neutron.NeutronAPI.get_neutron_network')
    @patch('zun.compute.api.API.network_attach')
    @patch('zun.objects.Container.get_by_uuid')
    def test_network_attach(self, mock_by_uuid, mock_attach, mock_get_network):
        built_requested_network = {
            'network': 'fake-net-id',
            'port': '',
            'router:external': False,
            'shared': False,
            'fixed_ip': '',
            'preserve_on_delete': False}
        query = 'network=private'
        self._test_network_attach(mock_by_uuid, mock_attach, mock_get_network,
                                  query, built_requested_network)
        mock_get_network.assert_called_once_with('private')

    @patch('zun.network.neutron.NeutronAPI.get_neutron_network')
    @patch('zun.compute.api.API.network_attach')
    @patch('zun.objects.Container.get_by_uuid')
    def test_network_attach_with_fixed_ip(self, mock_by_uuid, mock_attach,
                                          mock_get_network):
        built_requested_network = {
            'network': 'fake-net-id',
            'port': '',
            'router:external': False,
            'shared': False,
            'fixed_ip': '10.0.0.3',
            'preserve_on_delete': False}
        query = 'network=private&fixed_ip=10.0.0.3'
        self._test_network_attach(mock_by_uuid, mock_attach, mock_get_network,
                                  query, built_requested_network)
        mock_get_network.assert_called_once_with('private')

    @patch('zun.network.neutron.NeutronAPI.get_neutron_port')
    @patch('zun.network.neutron.NeutronAPI.ensure_neutron_port_usable')
    @patch('zun.network.neutron.NeutronAPI.get_neutron_network')
    @patch('zun.compute.api.API.network_attach')
    @patch('zun.objects.Container.get_by_uuid')
    def test_network_attach_with_port(self, mock_by_uuid, mock_attach,
                                      mock_get_network,
                                      mock_ensure, mock_get_port):
        mock_get_port.return_value = {
            'id': 'fake-port-id',
            'network_id': 'fake-net-id',
        }
        built_requested_network = {
            'network': 'fake-net-id',
            'port': 'fake-port-id',
            'router:external': False,
            'shared': False,
            'fixed_ip': '',
            'preserve_on_delete': True}
        query = 'port=fake-port'
        self._test_network_attach(mock_by_uuid, mock_attach, mock_get_network,
                                  query, built_requested_network)
        mock_get_port.assert_called_once_with('fake-port')
        mock_get_network.assert_called_once_with('fake-net-id')

    @patch('zun.network.neutron.NeutronAPI.get_available_network')
    @patch('zun.compute.api.API.network_attach')
    @patch('zun.objects.Container.get_by_uuid')
    def test_network_attach_without_params(self, mock_by_uuid, mock_attach,
                                           mock_get_network):
        built_requested_network = {
            'network': 'fake-net-id',
            'port': '',
            'router:external': False,
            'shared': False,
            'fixed_ip': '',
            'preserve_on_delete': False}
        query = ''
        self._test_network_attach(mock_by_uuid, mock_attach, mock_get_network,
                                  query, built_requested_network)
        mock_get_network.assert_called_once_with()

    def _test_network_attach(self, mock_by_uuid, mock_attach, mock_get_network,
                             query, built_requested_network):
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_by_uuid.return_value = test_container_obj
        container_uuid = test_container.get('uuid')
        fake_network = {'id': 'fake-net-id',
                        'router:external': False,
                        'shared': False}
        mock_get_network.return_value = fake_network
        url = '/v1/containers/%s/%s?%s' % (container_uuid, 'network_attach',
                                           query)
        response = self.post(url)
        self.assertEqual(200, response.status_int)
        mock_attach.assert_called_once_with(mock.ANY, test_container_obj,
                                            built_requested_network)

    @patch('zun.network.neutron.NeutronAPI.get_neutron_network')
    @patch('zun.compute.api.API.network_detach')
    @patch('zun.objects.Container.get_by_uuid')
    def test_network_detach(self, mock_by_uuid, mock_detach, mock_get_network):
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_by_uuid.return_value = test_container_obj
        container_uuid = test_container.get('uuid')
        mock_get_network.return_value = {'id': 'private'}
        mock_detach.return_value = None
        url = '/v1/containers/%s/%s?network=%s' % (container_uuid,
                                                   'network_detach',
                                                   'private')
        response = self.post(url)
        self.assertEqual(202, response.status_int)
        mock_detach.assert_called_once_with(mock.ANY, test_container_obj,
                                            'private')

    @patch('zun.network.neutron.NeutronAPI.get_neutron_port')
    @patch('zun.compute.api.API.network_detach')
    @patch('zun.objects.Container.get_by_uuid')
    def test_network_detach_with_port(self, mock_by_uuid, mock_detach,
                                      mock_get_port):
        port_uuid = uuidutils.generate_uuid()
        network_uuid = uuidutils.generate_uuid()
        test_container = utils.get_test_container(port=port_uuid,
                                                  network=network_uuid)
        test_container_obj = objects.Container(self.context, **test_container)
        mock_by_uuid.return_value = test_container_obj
        container_uuid = test_container.get('uuid')
        mock_get_port.return_value = {'network_id': network_uuid,
                                      'id': port_uuid}
        mock_detach.return_value = None
        url = '/v1/containers/%s/%s?port=%s' % (container_uuid,
                                                'network_detach',
                                                'fake-port')
        response = self.post(url)
        self.assertEqual(202, response.status_int)
        mock_detach.assert_called_once_with(mock.ANY, test_container_obj,
                                            network_uuid)

    @patch('zun.objects.Container.get_by_uuid')
    def test_network_list(self, mock_container_get_by_uuid):
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        mock_container_get_by_uuid.return_value = test_container_obj
        container_uuid = test_container.get('uuid')
        url = '/v1/containers/%s/%s' % (container_uuid, 'network_list')
        response = self.get(url)
        self.assertEqual(200, response.status_int)
        mock_container_get_by_uuid.assert_called_once_with(
            mock.ANY,
            test_container['uuid'])
        self._assert_networks(test_container['addresses'],
                              response.json['networks'])

    def _assert_networks(self, addresses, networks):
        self.assertEqual(len(addresses), len(networks))
        for network in networks:
            address_list = addresses[network['net_id']]
            self.assertEqual(len(address_list), len(network['fixed_ips']))
            for address in address_list:
                matched = 0
                for fixed_ip in network['fixed_ips']:
                    if (address['addr'] == fixed_ip['ip_address'] and
                            address['version'] == fixed_ip['version'] and
                            address['subnet_id'] == fixed_ip['subnet_id'] and
                            address['port'] == network['port_id']):
                        matched += 1
                self.assertEqual(1, matched)

    @mock.patch('zun.compute.api.API.remove_security_group')
    @mock.patch('zun.network.neutron.NeutronAPI.find_resourceid_by_name_or_id')
    @mock.patch('zun.api.utils.get_resource')
    def test_remove_security_group_by_uuid(self, mock_get_resource,
                                           mock_find_resourceid,
                                           mock_remove_security_group):
        headers = {"OpenStack-API-Version": "container 1.14"}
        test_container = utils.get_test_container(
            security_groups=['affb9021-964d-4b1b-80a8-9b9db60497e4'])
        test_container_obj = objects.Container(self.context, **test_container)
        mock_get_resource.return_value = test_container_obj
        mock_find_resourceid.return_value = \
            test_container_obj.security_groups[0]
        container_name = test_container.get('name')
        security_group_id_to_remove = test_container_obj.security_groups[0]
        url = '/v1/containers/%s/%s?name=%s' % (container_name,
                                                'remove_security_group',
                                                security_group_id_to_remove)
        response = self.post(url, headers=headers)
        self.assertEqual(202, response.status_int)
        self.assertEqual('application/json', response.content_type)
        mock_find_resourceid.assert_called_once_with(
            'security_group', security_group_id_to_remove, mock.ANY)
        mock_remove_security_group.assert_called_once_with(
            mock.ANY, test_container_obj,
            test_container_obj.security_groups[0])


class TestContainerEnforcement(api_base.FunctionalTest):

    def _common_policy_check(self, rule, func, *arg, **kwarg):
        rules = dict({rule: 'project_id:non_fake'},
                     **kwarg.pop('bypass_rules', {}))
        self.policy.set_rules(rules)
        response = func(*arg, **kwarg)
        self.assertEqual(403, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(
            "Policy doesn't allow %s to be performed." % rule,
            response.json['errors'][0]['detail'])

    def test_policy_disallow_get_all(self):
        self._common_policy_check(
            'container:get_all', self.get, '/v1/containers/',
            expect_errors=True)

    def test_policy_disallow_get_all_all_projects(self):
        self._common_policy_check(
            'container:get_all_all_projects',
            self.get, '/v1/containers/?all_projects=1',
            expect_errors=True,
            bypass_rules={'container:get_all': 'project_id:fake_project'})

    def test_policy_disallow_get_one(self):
        container = obj_utils.create_test_container(self.context)
        self._common_policy_check(
            'container:get_one', self.get,
            '/v1/containers/%s/' % container.uuid,
            expect_errors=True)

    def test_policy_disallow_get_one_all_projects(self):
        container = obj_utils.create_test_container(self.context)
        self._common_policy_check(
            'container:get_one_all_projects', self.get,
            '/v1/containers/%s/?all_projects=1' % container.uuid,
            expect_errors=True)

    def test_policy_disallow_update(self):
        container = obj_utils.create_test_container(self.context)
        params = {'cpu': 1}
        self._common_policy_check(
            'container:update', self.patch_json,
            '/containers/%s/' % container.uuid, params,
            expect_errors=True)

    def test_policy_disallow_create(self):
        params = ('{"name": "MyDocker", "image": "ubuntu",'
                  '"command": ["env"], "memory": "512"}')

        self._common_policy_check(
            'container:create', self.post, '/v1/containers/',
            params=params,
            content_type='application/json',
            expect_errors=True)

    def test_policy_disallow_delete(self):
        container = obj_utils.create_test_container(self.context)
        self._common_policy_check(
            'container:delete', self.delete,
            '/v1/containers/%s/' % container.uuid,
            expect_errors=True)

    def test_policy_disallow_delete_all_projects(self):
        container = obj_utils.create_test_container(self.context)
        self._common_policy_check(
            'container:delete_all_projects', self.delete,
            '/v1/containers/%s/?all_projects=1' % container.uuid,
            expect_errors=True)

    def test_policy_disallow_delete_force(self):
        container = obj_utils.create_test_container(self.context)
        self._common_policy_check(
            'container:delete_force', self.delete,
            '/v1/containers/%s/?force=True' % container.uuid,
            expect_errors=True,
            bypass_rules={'container:delete': 'project_id:fake_project'})

    def _owner_check(self, rule, func, *args, **kwargs):
        self.policy.set_rules({rule: "user_id:%(user_id)s"})
        response = func(*args, **kwargs)
        self.assertEqual(403, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(
            "Policy doesn't allow %s to be performed." % rule,
            response.json['errors'][0]['detail'])

    def test_policy_only_owner_get_one(self):
        container = obj_utils.create_test_container(self.context,
                                                    user_id='another')
        self._owner_check("container:get_one", self.get_json,
                          '/containers/%s/' % container.uuid,
                          expect_errors=True)

    def test_policy_only_owner_update(self):
        container = obj_utils.create_test_container(self.context,
                                                    user_id='another')
        self._owner_check(
            "container:update", self.patch_json,
            '/containers/%s/' % container.uuid,
            {'cpu': 1},
            expect_errors=True)

    def test_policy_only_owner_delete(self):
        container = obj_utils.create_test_container(self.context,
                                                    user_id='another')
        self._owner_check(
            "container:delete", self.delete,
            '/containers/%s/' % container.uuid,
            expect_errors=True)

    def test_policy_only_owner_logs(self):
        container = obj_utils.create_test_container(self.context,
                                                    user_id='another')
        self._owner_check("container:logs", self.get_json,
                          '/containers/%s/logs/' % container.uuid,
                          expect_errors=True)

    def test_policy_only_owner_execute(self):
        container = obj_utils.create_test_container(self.context,
                                                    user_id='another')
        self._owner_check("container:execute", self.post_json,
                          '/containers/%s/execute/' % container.uuid,
                          params={'command': 'ls'}, expect_errors=True)

    def test_policy_only_owner_actions(self):
        actions = ['start', 'stop', 'reboot', 'pause', 'unpause']
        container = obj_utils.create_test_container(self.context,
                                                    user_id='another')
        for action in actions:
            self._owner_check('container:%s' % action, self.post_json,
                              '/containers/%s/%s/' % (container.uuid, action),
                              {}, expect_errors=True)


class TestContainerActionController(api_base.FunctionalTest):

    def _format_action(self, action, expect_traceback=True):
        '''Remove keys that aren't serialized.'''
        to_delete = ('id', 'finish_time', 'created_at', 'updated_at',
                     'deleted_at', 'deleted')
        for key in to_delete:
            if key in action:
                del (action[key])
        for event in action.get('events', []):
            self._format_event(event, expect_traceback)
        return action

    def _format_event(self, event, expect_traceback=True):
        '''Remove keys that aren't serialized.'''
        to_delete = ['id', 'created_at', 'updated_at', 'deleted_at', 'deleted',
                     'action_id']
        if not expect_traceback:
            event['traceback'] = None
        for key in to_delete:
            if key in event:
                del (event[key])
        return event

    @mock.patch('zun.objects.Container.get_by_uuid')
    @mock.patch('zun.objects.ContainerAction.get_by_container_uuid')
    def test_list_actions(self, mock_get_by_container_uuid,
                          mock_container_get_by_uuid):
        test_container = utils.get_test_container()
        test_action = utils.get_test_action_value(
            container_uuid=test_container['uuid'])

        container_object = objects.Container(self.context, **test_container)
        action_object = objects.ContainerAction(self.context, **test_action)

        mock_container_get_by_uuid.return_value = container_object
        mock_get_by_container_uuid.return_value = [action_object]
        response = self.get('/v1/containers/%s/container_actions' %
                            test_container['uuid'])

        mock_get_by_container_uuid.assert_called_once_with(
            mock.ANY,
            test_container['uuid'])

        self.assertEqual(200, response.status_int)
        self.assertEqual(
            self._format_action(test_action),
            self._format_action(response.json['containerActions'][0]))

    @mock.patch('zun.objects.Container.get_by_uuid')
    @mock.patch('zun.common.policy.enforce')
    @mock.patch('zun.objects.ContainerActionEvent.get_by_action')
    @mock.patch('zun.objects.ContainerAction.get_by_request_id')
    def test_get_action_with_events_allowed(self, mock_get_by_request_id,
                                            mock_get_by_action, mock_policy,
                                            mock_container_get_by_uuid):
        mock_policy.return_value = True
        test_container = utils.get_test_container()
        test_action = utils.get_test_action_value(
            container_uuid=test_container['uuid'])
        test_event = utils.get_test_action_event_value(
            action_id=test_action['id'])
        test_action['events'] = [test_event]

        container_object = objects.Container(self.context, **test_container)
        action_object = objects.ContainerAction(self.context, **test_action)
        event_object = objects.ContainerActionEvent(self.context, **test_event)

        mock_container_get_by_uuid.return_value = container_object
        mock_get_by_request_id.return_value = action_object
        mock_get_by_action.return_value = [event_object]

        response = self.get('/v1/containers/%s/container_actions/%s' % (
            test_container['uuid'], test_action['request_id']))

        mock_get_by_request_id.assert_called_once_with(
            mock.ANY, test_container['uuid'], test_action['request_id'])
        mock_get_by_action.assert_called_once_with(mock.ANY, test_action['id'])

        self.assertEqual(200, response.status_int)
        self.assertEqual(self._format_action(test_action),
                         self._format_action(response.json))

    @mock.patch('zun.objects.Container.get_by_uuid')
    @mock.patch('zun.common.policy.enforce')
    @mock.patch('zun.objects.ContainerActionEvent.get_by_action')
    @mock.patch('zun.objects.ContainerAction.get_by_request_id')
    def test_get_action_with_events_not_allowed(self, mock_get_by_request_id,
                                                mock_get_by_action,
                                                mock_policy,
                                                mock_container_get_by_uuid):
        mock_policy.return_value = False
        test_container = utils.get_test_container()
        container_obj = objects.Container(self.context, **test_container)
        test_action = utils.get_test_action_value(
            container_uuid=test_container['uuid'])
        test_event = utils.get_test_action_event_value(
            action_id=test_action['id'])
        test_action['events'] = [test_event]
        action_object = objects.ContainerAction(self.context, **test_action)
        event_object = objects.ContainerActionEvent(self.context, **test_event)

        mock_container_get_by_uuid.return_value = container_obj
        mock_get_by_request_id.return_value = action_object
        mock_get_by_action.return_value = [event_object]

        response = self.get('/v1/containers/%s/container_actions/%s' % (
            test_container['uuid'], test_action['request_id']))

        mock_get_by_request_id.assert_called_once_with(
            mock.ANY, test_container['uuid'], test_action['request_id'])
        mock_get_by_action.assert_called_once_with(mock.ANY, test_action['id'])

        self.assertEqual(200, response.status_int)
        self.assertEqual(self._format_action(test_action,
                                             expect_traceback=False),
                         self._format_action(response.json))

    @mock.patch('zun.objects.Container.get_by_uuid')
    @mock.patch('zun.objects.ContainerAction.get_by_request_id')
    def test_action_not_found(self, mock_get_by_request_id,
                              mock_container_get_by_uuid):

        test_container = utils.get_test_container()
        container_obj = objects.Container(self.context, **test_container)

        mock_container_get_by_uuid.return_value = container_obj
        mock_get_by_request_id.return_value = None

        fake_request_id = 'request'

        self.assertRaises(AppError, self.get,
                          ('/v1/containers/%s/container_actions/%s' %
                           (test_container['uuid'], fake_request_id)))
        mock_get_by_request_id.assert_called_once_with(
            mock.ANY, test_container['uuid'], fake_request_id)

    @mock.patch('zun.objects.Container.get_by_uuid')
    def test_container_not_found(self, mock_container_get_by_uuid):
        test_container = utils.get_test_container()

        self.assertRaises(AppError, self.get,
                          ('/v1/containers/%s/container_actions'
                           % test_container['uuid']))
        mock_container_get_by_uuid.assert_called_once_with(
            mock.ANY, test_container['uuid'])


class TestContainerActionEnforcement(api_base.FunctionalTest):

    def _common_policy_check(self, rule, func, *arg, **kwarg):
        rules = dict({rule: 'project_id:non_fake'},
                     **kwarg.pop('bypass_rules', {}))
        self.policy.set_rules(rules)
        response = func(*arg, **kwarg)
        self.assertEqual(403, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(
            "Policy doesn't allow %s to be performed." % rule,
            response.json['errors'][0]['detail'])

    def test_list_actions_disallow_by_project(self):
        container = obj_utils.create_test_container(self.context)

        self._common_policy_check(
            'container:actions', self.get,
            '/v1/containers/%s/container_actions/' % container.uuid,
            expect_errors=True)

    def test_get_action_disallow_by_project(self):
        container = obj_utils.create_test_container(self.context)

        self._common_policy_check(
            'container:actions', self.get,
            '/v1/containers/%s/container_actions/fake_request' %
            container.uuid, expect_errors=True)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import zun.conf

CONF = zun.conf.CONF


def basedir_def(*args):
    """Return an uninterpolated path relative to $pybasedir."""
    return os.path.join('$pybasedir', *args)


def bindir_def(*args):
    """Return an uninterpolated path relative to $bindir."""
    return os.path.join('$bindir', *args)


def state_path_def(*args):
    """Return an uninterpolated path relative to $state_path."""
    return os.path.join('$state_path', *args)


def basedir_rel(*args):
    """Return a path relative to $pybasedir."""
    return os.path.join(CONF.pybasedir, *args)


def bindir_rel(*args):
    """Return a path relative to $bindir."""
    return os.path.join(CONF.bindir, *args)


def state_path_rel(*args):
    """Return a path relative to $state_path."""
    return os.path.join(CONF.state_path, *args)



# Copyright 2013 - Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Zun base exception handling.

Includes decorator for re-raising Zun-type exceptions.

"""

import functools
import inspect
import re
import sys

from keystoneclient import exceptions as keystone_exceptions
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import uuidutils
import pecan
from webob import util as woutil

from zun.common.i18n import _
import zun.conf

LOG = logging.getLogger(__name__)

CONF = zun.conf.CONF

try:
    CONF.import_opt('fatal_exception_format_errors',
                    'oslo_versionedobjects.exception')
except cfg.NoSuchOptError:
    # Note:work around for zun run against master branch
    # in devstack gate job, as zun not branched yet
    # versionobjects kilo/master different version can
    # cause issue here. As it changed import group. So
    # add here before branch to prevent gate failure.
    # Bug: #1447873
    CONF.import_opt('fatal_exception_format_errors',
                    'oslo_versionedobjects.exception',
                    group='oslo_versionedobjects')


def wrap_exception(notifier=None, event_type=None):
    """This decorator wraps a method to catch any exceptions.

    It logs the exception as well as optionally sending
    it to the notification system.
    """
    def inner(f):
        def wrapped(self, context, *args, **kwargs):
            # Don't store self or context in the payload, it now seems to
            # contain confidential information.
            try:
                return f(self, context, *args, **kwargs)
            except Exception as e:
                with excutils.save_and_reraise_exception():
                    if notifier:
                        call_dict = inspect.getcallargs(f, self, context,
                                                        *args, **kwargs)
                        payload = dict(exception=e,
                                       private=dict(args=call_dict)
                                       )

                        temp_type = event_type
                        if not temp_type:
                            # If f has multiple decorators, they must use
                            # functools.wraps to ensure the name is
                            # propagated.
                            temp_type = f.__name__

                        notifier.error(context, temp_type, payload)

        return functools.wraps(f)(wrapped)
    return inner


OBFUSCATED_MSG = _('Your request could not be handled '
                   'because of a problem in the server. '
                   'Error Correlation id is: %s')


def wrap_controller_exception(func, func_server_error, func_client_error):
    """This decorator wraps controllers methods to handle exceptions:

    - if an unhandled Exception or a ZunException with an error code >=500
      is catched, raise a http 5xx ClientSideError and correlates it with a log
      message

    - if a ZunException is catched and its error code is <500, raise a http
      4xx and logs the excp in debug mode

    """
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as excp:
            if isinstance(excp, ZunException):
                http_error_code = excp.code
            else:
                http_error_code = 500

            if http_error_code >= 500:
                # log the error message with its associated
                # correlation id
                log_correlation_id = uuidutils.generate_uuid()
                LOG.exception("%(correlation_id)s:%(excp)s",
                              {'correlation_id': log_correlation_id,
                               'excp': str(excp)})
                # raise a client error with an obfuscated message
                return func_server_error(log_correlation_id, http_error_code)
            else:
                # raise a client error the original message
                LOG.debug(excp)
                return func_client_error(excp, http_error_code)
    return wrapped


def convert_excp_to_err_code(excp_name):
    """Convert Exception class name (CamelCase) to error-code (Snake-case)"""
    words = re.findall(r'[A-Z]?[a-z]+|[A-Z]{2,}(?=[A-Z][a-z]|\d|\W|$)|\d+',
                       excp_name)
    return '-'.join([str.lower(word) for word in words])


def wrap_pecan_controller_exception(func):
    """This decorator wraps pecan controllers to handle exceptions."""
    def _func_server_error(log_correlation_id, status_code):
        pecan.response.status = status_code
        return {
            'faultcode': 'Server',
            'status_code': status_code,
            'title': woutil.status_reasons[status_code],
            'description': str(OBFUSCATED_MSG % log_correlation_id),
        }

    def _func_client_error(excp, status_code):
        pecan.response.status = status_code
        return {
            'faultcode': 'Client',
            'faultstring': convert_excp_to_err_code(excp.__class__.__name__),
            'status_code': status_code,
            'title': str(excp),
            'description': str(excp),
        }

    return wrap_controller_exception(func,
                                     _func_server_error,
                                     _func_client_error)


def wrap_keystone_exception(func):
    """Wrap keystone exceptions and throw Zun specific exceptions."""
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except keystone_exceptions.AuthorizationFailure:
            raise AuthorizationFailure(
                client=func.__name__, message="reason: %s" % sys.exc_info()[1])
        except keystone_exceptions.ClientException:
            raise AuthorizationFailure(
                client=func.__name__,
                message="unexpected keystone client error occurred: %s"
                        % sys.exc_info()[1])
    return wrapped


class ZunException(Exception):
    """Base Zun Exception

    To correctly use this class, inherit from it and define
    a 'message' property. That message will get printf'd
    with the keyword arguments provided to the constructor.

    """
    message = _("An unknown exception occurred.")
    code = 500

    def __init__(self, message=None, **kwargs):
        self.kwargs = kwargs

        if 'code' not in self.kwargs and hasattr(self, 'code'):
            self.kwargs['code'] = self.code

        if message:
            self.message = message

        try:
            self.message = self.message % kwargs
        except KeyError:
            # kwargs doesn't match a variable in the message
            # log the issue and the kwargs
            LOG.exception('Exception in string format operation, '
                          'kwargs: %s', kwargs)
            try:
                ferr = CONF.fatal_exception_format_errors
            except cfg.NoSuchOptError:
                ferr = CONF.oslo_versionedobjects.fatal_exception_format_errors
            if ferr:
                raise

        super(ZunException, self).__init__(self.message)

    def __str__(self):
        return self.message

    def format_message(self):
        if self.__class__.__name__.endswith('_Remote'):
            return self.args[0]
        else:
            return str(self)


class ObjectNotFound(ZunException):
    message = _("The %(name)s %(id)s could not be found.")


class ObjectNotUnique(ZunException):
    message = _("The %(name)s already exists.")


class ObjectActionError(ZunException):
    message = _('Object action %(action)s failed because: %(reason)s')


class ResourceNotFound(ObjectNotFound):
    message = _("The %(name)s resource %(id)s could not be found.")
    code = 404


class ResourceExists(ObjectNotUnique):
    message = _("The %(name)s resource already exists.")
    code = 409


class AuthorizationFailure(ZunException):
    message = _("%(client)s connection failed. %(message)s")


class UnsupportedObjectError(ZunException):
    message = _('Unsupported object type %(objtype)s')


class IncompatibleObjectVersion(ZunException):
    message = _('Version %(objver)s of %(objname)s is not supported')


class OrphanedObjectError(ZunException):
    message = _('Cannot call %(method)s on orphaned %(objtype)s object')


class Invalid(ZunException):
    message = _("Unacceptable parameters.")
    code = 400


class InvalidValue(Invalid):
    message = _("Received value '%(value)s' is invalid for type %(type)s.")


class ValidationError(Invalid):
    message = "%(detail)s"


class SchemaValidationError(ValidationError):
    message = "%(detail)s"


class InvalidUUID(Invalid):
    message = _("Expected a uuid but received %(uuid)s.")


class InvalidName(Invalid):
    message = _("Expected a name but received %(uuid)s.")


class InvalidDiscoveryURL(Invalid):
    message = _("Received invalid discovery URL '%(discovery_url)s' for "
                "discovery endpoint '%(discovery_endpoint)s'.")


class GetDiscoveryUrlFailed(ZunException):
    message = _("Failed to get discovery url from '%(discovery_endpoint)s'.")


class InvalidUuidOrName(Invalid):
    message = _("Expected a name or uuid but received %(uuid)s.")


class InvalidIdentity(Invalid):
    message = _("Expected an uuid or int but received %(identity)s.")


class InvalidCsr(Invalid):
    message = _("Received invalid csr %(csr)s.")


class HTTPNotFound(ResourceNotFound):
    pass


class Conflict(ZunException):
    message = _('Conflict.')
    code = 409


class ConflictOptions(Conflict):
    message = _('Conflicting options.')


class InvalidState(Conflict):
    message = _("Invalid resource state.")


# Cannot be templated as the error syntax varies.
# msg needs to be constructed when raised.
class InvalidParameterValue(Invalid):
    message = _("%(err)s")


class InvalidParamInVersion(Invalid):
    message = _('Invalid param %(param)s because current request '
                'version is %(req_version)s. %(param)s is only '
                'supported from version %(min_version)s')


class InvalidQuotaValue(Invalid):
    message = _("Change would make usage less than 0 for the following "
                "resources: %(unders)s")


class InvalidQuotaMethodUsage(Invalid):
    message = _("Wrong quota method %(method)s used on resource %(res)s")


class PatchError(Invalid):
    message = _("Couldn't apply patch '%(patch)s'. Reason: %(reason)s")


class NotAuthorized(ZunException):
    message = _("Not authorized.")
    code = 403


class ConfigInvalid(ZunException):
    message = _("Invalid configuration file. %(error_msg)s")


class PolicyNotAuthorized(NotAuthorized):
    message = _("Policy doesn't allow %(action)s to be performed.")


class ContainerNotFound(HTTPNotFound):
    message = _("Container %(container)s could not be found.")


class ContainerHostNotUp(ZunException):
    message = _("Container %(container)s host %(host)s is not up.")


class ComputeNodeNotFound(HTTPNotFound):
    message = _("Compute node %(compute_node)s could not be found.")


class NetworkNotFound(HTTPNotFound):
    message = _("Neutron network %(network)s could not be found.")


class RegistryNotFound(HTTPNotFound):
    message = _("Registry %(registry)s could not be found.")


class NetworkAlreadyExists(ResourceExists):
    message = _("A network with %(field)s %(value)s already exists.")

    def __init__(self, field, value):
        self.field = field
        self.value = value
        super(NetworkAlreadyExists, self).__init__(field=field, value=value)


class PortNotFound(HTTPNotFound):
    message = _("Neutron port %(port)s could not be found.")


class VolumeMappingNotFound(HTTPNotFound):
    message = _("Volume mapping %(volume_mapping)s could not be found.")


class VolumeNotFound(HTTPNotFound):
    message = _("Volume %(volume)s could not be found.")


class ImageNotFound(Invalid):
    message = _("Image %(image)s could not be found.")


class OperationNotSupported(ZunException):
    message = _("The operation is not supported.")


class ZunServiceNotFound(HTTPNotFound):
    message = _("Zun service %(binary)s on host %(host)s could not be found.")


class ResourceClassNotFound(HTTPNotFound):
    message = _("Resource class %(resource_class)s could not be found.")


class InventoryNotFound(HTTPNotFound):
    message = _("Inventory %(inventory)s could not be found.")


class AllocationNotFound(HTTPNotFound):
    message = _("Allocation %(allocation)s could not be found.")


class ContainerAlreadyExists(ResourceExists):
    message = _("A container with %(field)s %(value)s already exists.")


class ExecInstanceAlreadyExists(ResourceExists):
    message = _("An exec instance with exec_id %(exec_id)s already exists "
                "in container.")


class ComputeNodeAlreadyExists(ResourceExists):
    message = _("A compute node with %(field)s %(value)s already exists.")


class ImageAlreadyExists(ResourceExists):
    message = _("An image with tag %(tag)s and repo %(repo)s already exists.")


class ZunServiceAlreadyExists(ResourceExists):
    message = _("Service %(binary)s on host %(host)s already exists.")


class ResourceProviderAlreadyExists(ResourceExists):
    message = _("A resource provider with %(field)s %(value)s already exists.")


class ResourceClassAlreadyExists(ResourceExists):
    message = _("A resource class with %(field)s %(value)s already exists.")


class VolumeMappingAlreadyExists(ResourceExists):
    message = _("A volume mapping with %(field)s %(value)s already exists.")


class VolumeAlreadyExists(ResourceExists):
    message = _("A volume with %(field)s %(value)s already exists.")


class RegistryAlreadyExists(ResourceExists):
    message = _("A registry with %(field)s %(value)s already exists.")


class PortNotUsable(Invalid):
    message = _("Port %(port)s not usable for the container.")


class PortInUse(Invalid):
    message = _("Port %(port)s is still in use.")


class VolumeNotUsable(Invalid):
    message = _("Volume %(volume)s not usable for the container.")


class VolumeInUse(Invalid):
    message = _("Volume %(volume)s is still in use.")


class PortBindingFailed(Invalid):
    message = _("Binding failed for port %(port)s, please check neutron "
                "logs for more information.")


class UniqueConstraintViolated(ResourceExists):
    message = _("A resource with %(fields)s violates unique constraint.")


class InvalidStateException(ZunException):
    message = _("Cannot %(action)s container %(id)s in %(actual_state)s state")
    code = 409


class DockerError(ZunException):
    message = _("Docker internal error: %(error_msg)s.")


class PollTimeOut(ZunException):
    message = _("Polling request timed out.")


class ServerInError(ZunException):
    message = _('Went to status %(resource_status)s due to '
                '"%(status_reason)s"')


class ServerUnknownStatus(ZunException):
    message = _('%(result)s - Unknown status %(resource_status)s due to '
                '"%(status_reason)s"')


class EntityNotFound(ZunException):
    message = _("The %(entity)s (%(name)s) could not be found.")


class CommandError(ZunException):
    message = _("The command: %(cmd)s failed on the system, due to %(error)s")


class NoValidHost(ZunException):
    message = _("No valid host was found. %(reason)s")


class NoInteractiveFlag(Invalid):
    message = _("%(msg)s")


class CPUPinningUnknown(ZunException):
    message = _("CPU set to pin %(requested)s must be a subset of "
                "known CPU set %(cpuset)s")


class CPUUnpinningUnknown(Invalid):
    message = _("CPU set to unpin %(requested)s must be a subset of "
                "known CPU set %(cpuset)s")


class CPUPinningInvalid(Invalid):
    message = _("CPU set to pin %(requested)s must be a subset of "
                "free CPU set %(free)s")


class CPUUnpinningInvalid(Invalid):
    message = _("CPU set to unpin %(requested)s must be a subset of "
                "pinned CPU set %(pinned)s")


class NotFound(ZunException):
    message = _("Resource could not be found.")
    code = 404


class SchedulerHostFilterNotFound(NotFound):
    message = _("Scheduler Host Filter %(filter_name)s could not be found.")


class ClassNotFound(NotFound):
    message = _("Class %(class_name)s could not be found: %(exception)s")


class ApiVersionsIntersect(ZunException):
    message = _("Version of %(name)s %(min_ver)s %(max_ver)s intersects "
                "with another versions.")


class ConnectionFailed(ZunException):
    message = _("Failed to connect to remote host")


class SocketException(ZunException):
    message = _("Socket exceptions")


class InvalidWebsocketUrl(ZunException):
    message = _("Websocket Url invalid")


class InvalidWebsocketToken(ZunException):
    message = _("Websocket token is invalid")


class ResourcesUnavailable(ZunException):
    message = _("Insufficient compute resources: %(reason)s.")


class MakeFileSystemException(ZunException):
    message = _("Unexpected error while make file system")


class MountException(ZunException):
    message = _("Unexpected error while mount device")


class UnmountException(ZunException):
    message = _("Unexpected error while do umount")


class FileNotFound(ZunException):
    message = _("The expected file not exist")


class PciConfigInvalidWhitelist(ZunException):
    message = _("Invalid PCI devices Whitelist config %(reason)s")


class PciDeviceWrongAddressFormat(ZunException):
    message = _("The PCI address %(address)s has an incorrect format.")


class PciDeviceInvalidDeviceName(ZunException):
    message = _("Invalid PCI Whitelist: "
                "The PCI whitelist can specify devname or address,"
                " but not both")


class PciDeviceNotFoundById(NotFound):
    message = _("PCI device %(id)s not found")


class PciDeviceNotFound(NotFound):
    message = _("PCI Device %(node_id)s:%(address)s not found.")


class PciDevicePoolEmpty(ZunException):
    message = _(
        "Attempt to consume PCI device %(compute_node_uuid)s:%(address)s "
        "from empty pool")


class CapsuleAlreadyExists(ResourceExists):
    message = _("A capsule with %(field)s %(value)s already exists.")


class CapsuleNotFound(HTTPNotFound):
    message = _("Capsule %(capsule)s could not be found.")


class InvalidCapsuleTemplate(ZunException):
    message = _("Invalid capsule template: %(reason)s.")


class FailedParseStringToJson(ZunException):
    message = _("Failed parse string to json: %(reason)s.")


class ExternalNetworkAttachForbidden(NotAuthorized):
    message = _("It is not allowed to create an interface on "
                "external network %(network_uuid)s")


class PciDeviceInvalidStatus(Invalid):
    message = _(
        "PCI device %(compute_node_id)s:%(address)s is %(status)s "
        "instead of %(hopestatus)s")


class PciDeviceVFInvalidStatus(Invalid):
    message = _(
        "Not all Virtual Functions of PF %(compute_node_id)s:%(address)s "
        "are free.")


class PciDevicePFInvalidStatus(Invalid):
    message = _(
        "Physical Function %(compute_node_id)s:%(address)s, related to VF"
        " %(compute_node_id)s:%(vf_address)s is %(status)s "
        "instead of %(hopestatus)s")


class PciDeviceInvalidOwner(Invalid):
    message = _(
        "PCI device %(compute_node_id)s:%(address)s is owned by %(owner)s "
        "instead of %(hopeowner)s")


class VolumeCreateFailed(Invalid):
    message = _("Volume Creation failed: %(creation_failed)s")


class VolumeDeleteFailed(Invalid):
    message = _("Volume Deletion failed: %(deletion_failed)s")


class SecurityGroupCannotBeApplied(Invalid):
    message = _("Security groups cannot apply to the container.")


class SecurityGroupCannotBeRemoved(Invalid):
    message = _("Security groups cannot be removed from the container.")


class ContainerActionNotFound(ZunException):
    message = _("Action for request_id %(request_id)s on container"
                " %(container_uuid)s not fount")


class ContainerActionEventNotFound(ZunException):
    message = _("Event %(event)s not found for action id %(action_id)s")


class OverQuota(ZunException):
    message = _("Quota exceeded for resources: %(overs)s")
    code = 403


class QuotaNotFound(NotFound):
    message = _("Quota could not be found.")


class QuotaUsageNotFound(NotFound):
    message = _("Quota usage could not be found")


class ProjectQuotaNotFound(QuotaNotFound):
    message = _("Quota for project %(project_id)s could not be found.")


class QuotaAlreadyExists(ZunException):
    message = _("Quota already exists for project %(project_id)s, "
                "resource %(resource)s.")


class QuotaClassNotFound(QuotaNotFound):
    message = _("Quota class %(class_name)s could not be found.")


class QuotaResourceUnknown(QuotaNotFound):
    message = _("Unknow quota resources %(unknown)s.")


class Base64Exception(Invalid):
    message = _("Invalid Base 64 file data")


class InvalidEncryptionKey(ZunException):
    message = _('Can not decrypt data with the auth_encryption_key '
                'in zun config.')


class InvalidReference(ZunException):
    message = _('invalid reference')


class ReferenceInvalidFormat(InvalidReference):
    message = _('invalid reference format')


class NameEmpty(InvalidReference):
    message = _('repository name must have at least one component')


class NameTooLong(InvalidReference):
    message = _('repository name must not be more than %(length_max)s '
                'characters')


# An exception with this name is used on both sides of the placement/zun
# interaction.
class ResourceProviderInUse(ZunException):
    message = _("Resource provider has allocations.")


class ResourceProviderRetrievalFailed(ZunException):
    message = _("Failed to get resource provider with UUID %(uuid)s")


class ResourceProviderAggregateRetrievalFailed(ZunException):
    message = _("Failed to get aggregates for resource provider with UUID"
                " %(uuid)s")


class ResourceProviderTraitRetrievalFailed(ZunException):
    message = _("Failed to get traits for resource provider with UUID"
                " %(uuid)s")


class ResourceProviderCreationFailed(ZunException):
    message = _("Failed to create resource provider %(name)s")


class ResourceProviderDeletionFailed(ZunException):
    message = _("Failed to delete resource provider %(uuid)s")


class ResourceProviderUpdateFailed(ZunException):
    message = _("Failed to update resource provider via URL %(url)s: "
                "%(error)s")


class ResourceProviderNotFound(NotFound):
    message = _("No such resource provider %(name_or_uuid)s.")


class ResourceProviderSyncFailed(ZunException):
    message = _("Failed to synchronize the placement service with resource "
                "provider information supplied by the compute host.")


class PlacementAPIConnectFailure(ZunException):
    message = _("Unable to communicate with the Placement API.")


class PlacementAPIConflict(ZunException):
    """Any 409 error from placement APIs should use (a subclass of) this
    exception.
    """
    message = _("A conflict was encountered attempting to invoke the "
                "placement API at URL %(url)s: %(error)s")


class ResourceProviderUpdateConflict(PlacementAPIConflict):
    """A 409 caused by generation mismatch from attempting to update an
    existing provider record or its associated data (aggregates, traits, etc.).
    """
    message = _("A conflict was encountered attempting to update resource "
                "provider %(uuid)s (generation %(generation)d): %(error)s")


class InvalidResourceClass(Invalid):
    message = _("Resource class '%(resource_class)s' invalid.")


class InvalidResourceAmount(Invalid):
    message = _("Resource amounts must be integers. Received '%(amount)s'.")


class InvalidInventory(Invalid):
    message = _("Inventory for '%(resource_class)s' on "
                "resource provider '%(resource_provider)s' invalid.")


# An exception with this name is used on both sides of the placement/
# nova interaction.
class InventoryInUse(InvalidInventory):
    pass


class UsagesRetrievalFailed(ZunException):
    message = _("Failed to retrieve usages for project '%(project_id)s' and "
                "user '%(user_id)s'.")


class AllocationUpdateFailed(ZunException):
    message = _('Failed to update allocations for consumer %(consumer_uuid)s. '
                'Error: %(error)s')


class ConsumerAllocationRetrievalFailed(ZunException):
    message = _("Failed to retrieve allocations for consumer "
                "%(consumer_uuid)s: %(error)s")


class TraitRetrievalFailed(ZunException):
    message = _("Failed to get traits for resource provider with UUID"
                " %(uuid)s")


class TraitCreationFailed(ZunException):
    message = _("Failed to create trait %(name)s: %(error)s")


class AllocationMoveFailed(ZunException):
    message = _('Failed to move allocations from consumer %(source_consumer)s '
                'to consumer %(target_consumer)s. '
                'Error: %(error)s')


class ResourceProviderAllocationRetrievalFailed(ZunException):
    message = _("Failed to retrieve allocations for resource provider "
                "%(rp_uuid)s: %(error)s")


class ComputeHostNotFound(NotFound):
    message = _("Compute host %(host)s could not be found.")


class RequestedHostNotFound(NotFound):
    message = _("Requested host %(host)s could not be found.")


class CNIError(ZunException):
    pass


class ResourceNotReady(CNIError):
    message = _("Resource not ready: %(resource)s")
